# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# Local imports
from .const import MRT_TABLE_DUMP_V2_RIB_SPECIFIC_ANY
from .const import MRT_TABLE_DUMP_V2_RIB_GENERIC_ANY
from .const import MRT_TABLE_DUMP_V2_RIB_IPV6_ANY
from ..bgp.attr import unpack_mrt_bgp_attr
from ..bgp.nlri import unpack_mrt_bgp_nlri
from ..bgp.const import BGP_SAFI_UNICAST
from ..bgp.const import BGP_SAFI_MULTICAST
from ...util import CACHE_TS
from ...const import IPV4
from ...const import IPV6
from ...const import IPV4_STR
from ...const import IPV6_STR
from ...const import AF_INET
from ...const import AF_INET6
from ...const import AFI_IPV4
from ...const import AFI_IPV6
from ...const import STRUCT_2B
from ...const import STRUCT_4B
from ...const import STRUCT_8B8B
from ...const import UTF8
from ...const import DATETIME_FORMAT_USEC
from ...const import DATETIME_FORMAT_MIN
from ...const import struct_unpack
from ...const import socket_inet_ntop
from ...const import datetime_utcfromtimestamp
from ....model.attr import FTL_ATTR_BGP_PEER_TABLE_COLLECTOR_BGP_ID
from ....model.attr import FTL_ATTR_BGP_PEER_TABLE_VIEW_NAME
from ....model.attr import FTL_ATTR_BGP_PEER_TABLE_PEER_PROTOCOL
from ....model.attr import FTL_ATTR_BGP_PEER_TABLE_PEER_BGP_ID
from ....model.attr import FTL_ATTR_BGP_PEER_TABLE_PEER_AS
from ....model.attr import FTL_ATTR_BGP_PEER_TABLE_PEER_IP
from ....model.attr import FTL_ATTR_BGP_PEER_TABLE_COLLECTOR_BGP_ID_HUMAN
from ....model.attr import FTL_ATTR_BGP_PEER_TABLE_PEER_PROTOCOL_HUMAN
from ....model.attr import FTL_ATTR_BGP_PEER_TABLE_PEER_BGP_ID_HUMAN
from ....model.attr import FTL_ATTR_BGP_PEER_TABLE_PEER_IP_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_SOURCE
from ....model.attr import FTL_ATTR_BGP_ROUTE_SEQUENCE
from ....model.attr import FTL_ATTR_BGP_ROUTE_TIMESTAMP
from ....model.attr import FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL
from ....model.attr import FTL_ATTR_BGP_ROUTE_PEER_BGP_ID
from ....model.attr import FTL_ATTR_BGP_ROUTE_PEER_AS
from ....model.attr import FTL_ATTR_BGP_ROUTE_PEER_IP
from ....model.attr import FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL
from ....model.attr import FTL_ATTR_BGP_ROUTE_PREFIX
from ....model.attr import FTL_ATTR_BGP_ROUTE_PATH_ID
from ....model.attr import FTL_ATTR_BGP_ROUTE_SOURCE_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_TIMESTAMP_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_PEER_BGP_ID_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_PEER_IP_HUMAN
from ....model.attr import FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV4
from ....model.attr import FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV6
from ....model.attr import FTL_ATTR_BGP_STATS_MRT_FIXES
from ....model.attr import FTL_ATTR_BGP_STATS_MRT_FIXES_HUMAN
from ....model.const import FTL_ATTR_BGP_ROUTE_SOURCE_RIB
from ....model.const import FTL_ATTR_BGP_ROUTE_SOURCE_RIB_STR
from ....model.const import FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_ADDPATH
from ....model.const import FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_ADDPATH_STR
from ....model.const import FTL_ATTR_BGP_ERROR_REASON_MISSING_DATA
from ....model.const import FTL_ATTR_BGP_ERROR_REASON_INVALID_TYPE
from ....model.const import FTL_ATTR_BGP_ERROR_REASON_INVALID_PROTOCOL
from ....model.const import FTL_ATTR_BGP_ERROR_REASON_INVALID_PREFIX
from ....model.record import FTL_RECORD_BGP_PEER_TABLE
from ....model.record import FTL_RECORD_BGP_ROUTE
from ....model.record import FTL_RECORD_BGP_STATS
from ....model.error import FtlMrtHeaderError
from ....model.error import FtlMrtDataError


# pylint: disable-next=unused-argument
def unpack_mrt_entry_tdv2_index(caches, stats_record, peer_table_records, entry_bytes, peer_table):
    """ Parse MRT peer index table.
    """
    # Access peer table record template
    peer_table_init, peer_table_emit, _ = peer_table_records

    # Initialize peer table record view
    peer_table_record_view = list(peer_table_init)

    # -----------------------------------------
    # [RFC6396] 4.3.1. PEER_INDEX_TABLE Subtype
    # -----------------------------------------
    #  0                   1                   2                   3
    #  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                      Collector BGP ID                         |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |       View Name Length        |     View Name (variable)      |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |          Peer Count           |    Peer Entries (variable)
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    # Prepare byte offset
    offset = 0

    # Parse collector BGP ID (IPv4 only)
    if FTL_ATTR_BGP_PEER_TABLE_COLLECTOR_BGP_ID >= 0:
        col_bgp_id = entry_bytes[offset:offset + 4]
        if FTL_ATTR_BGP_PEER_TABLE_COLLECTOR_BGP_ID_HUMAN:
            peer_table_record_view[FTL_ATTR_BGP_PEER_TABLE_COLLECTOR_BGP_ID] = socket_inet_ntop(AF_INET, col_bgp_id)
        else:
            peer_table_record_view[FTL_ATTR_BGP_PEER_TABLE_COLLECTOR_BGP_ID] = struct_unpack(STRUCT_4B, col_bgp_id)[0]
    offset += 4

    # Parse view name length
    view_name_length = struct_unpack(STRUCT_2B, entry_bytes[offset:offset + 2])[0]
    offset += 2

    # Parse view name
    if view_name_length > 0:
        if FTL_ATTR_BGP_PEER_TABLE_VIEW_NAME >= 0:
            view_name = bytes(entry_bytes[offset:offset + view_name_length]).decode(UTF8)
            peer_table_record_view[FTL_ATTR_BGP_PEER_TABLE_VIEW_NAME] = view_name
        offset += view_name_length

    # Parse peer count
    peer_count = struct_unpack(STRUCT_2B, entry_bytes[offset:offset + 2])[0]
    offset += 2

    # Parse peer entries
    for _ in range(peer_count):

        # Initialize peer table record
        peer_table_record = list(peer_table_record_view)

        # -----------------------------------------
        # [RFC6396] 4.3.1. PEER_INDEX_TABLE Subtype
        # -----------------------------------------
        #  0                   1                   2                   3
        #  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # |   Peer Type   |
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # |                         Peer BGP ID                           |
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # |                   Peer IP Address (variable)                  |
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # |                        Peer AS (variable)                     |
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

        # Parse peer type
        peer_type = entry_bytes[offset]
        offset += 1

        # Parse peer protocol and AS length
        afinet, iplen = (AF_INET6, 16) if peer_type & 1 else (AF_INET, 4)
        asfmt, aslen = (STRUCT_4B, 4) if peer_type & 2 else (STRUCT_2B, 2)

        # Add peer protocol to peer table record
        peer_proto_int, peer_proto_str = (IPV6, IPV6_STR) if afinet == AF_INET6 else (IPV4, IPV4_STR)
        if FTL_ATTR_BGP_PEER_TABLE_PEER_PROTOCOL >= 0:
            peer_proto = peer_proto_str if FTL_ATTR_BGP_PEER_TABLE_PEER_PROTOCOL_HUMAN else peer_proto_int
            peer_table_record[FTL_ATTR_BGP_PEER_TABLE_PEER_PROTOCOL] = peer_proto

        # Add peer BGP ID to peer table record
        peer_bgp_id = entry_bytes[offset:offset + 4]
        peer_bgp_id_str = socket_inet_ntop(AF_INET, peer_bgp_id)
        peer_bgp_id_int = struct_unpack(STRUCT_4B, peer_bgp_id)[0]
        if FTL_ATTR_BGP_PEER_TABLE_PEER_BGP_ID >= 0:
            peer_bgp_id = peer_bgp_id_str if FTL_ATTR_BGP_PEER_TABLE_PEER_BGP_ID_HUMAN else peer_bgp_id_int
            peer_table_record[FTL_ATTR_BGP_PEER_TABLE_PEER_BGP_ID] = peer_bgp_id
        offset += 4

        # Add peer IP to peer table record
        peer_ip = entry_bytes[offset:offset + iplen]
        peer_ip_str = socket_inet_ntop(afinet, peer_ip)
        if afinet == AF_INET6:
            net, host = struct_unpack(STRUCT_8B8B, peer_ip)
            peer_ip_int = (net << 64) + host
        else:
            peer_ip_int = struct_unpack(STRUCT_4B, peer_ip)[0]
        if FTL_ATTR_BGP_PEER_TABLE_PEER_IP >= 0:
            peer_ip = peer_ip_str if FTL_ATTR_BGP_PEER_TABLE_PEER_IP_HUMAN else peer_ip_int
            peer_table_record[FTL_ATTR_BGP_PEER_TABLE_PEER_IP] = peer_ip
        offset += iplen

        # Add peer AS to peer table record
        peer_as = struct_unpack(asfmt, entry_bytes[offset:offset + aslen])[0]
        if FTL_ATTR_BGP_PEER_TABLE_PEER_AS >= 0:
            peer_table_record[FTL_ATTR_BGP_PEER_TABLE_PEER_AS] = peer_as
        offset += aslen

        # Update peer table
        peer_proto_route = peer_proto_str if FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL_HUMAN else peer_proto_int
        peer_ip_route = peer_ip_str if FTL_ATTR_BGP_ROUTE_PEER_IP_HUMAN else peer_ip_int
        peer_bgp_id_route = peer_bgp_id_str if FTL_ATTR_BGP_ROUTE_PEER_BGP_ID_HUMAN else peer_bgp_id_int
        peer_table.append((peer_proto_route, peer_bgp_id_route, peer_as, peer_ip_route))

        # Yield peer table record
        if FTL_RECORD_BGP_PEER_TABLE:
            yield peer_table_emit(peer_table_record)


def unpack_mrt_entry_tdv2_rib(caches, stats_record, route_records, entry_bytes, mrt_subtype, peer_table,
                              addpath=False, addpath_error=None):
    """ Parse MRT table dump v2.
    """
    # Prepare ADD-PATH RIB flag and protocol
    addpath_rib, afinet = addpath, None

    # Access route record template
    route_init, route_emit, route_error = route_records

    # Initialize route record prefix
    route_record_prefix = list(route_init)

    # Add source to route record prefix
    if FTL_ATTR_BGP_ROUTE_SOURCE >= 0:
        if FTL_ATTR_BGP_ROUTE_SOURCE_HUMAN:
            route_record_prefix[FTL_ATTR_BGP_ROUTE_SOURCE] = FTL_ATTR_BGP_ROUTE_SOURCE_RIB_STR
        else:
            route_record_prefix[FTL_ATTR_BGP_ROUTE_SOURCE] = FTL_ATTR_BGP_ROUTE_SOURCE_RIB

    # Prepare byte offset
    offset = 0

    # Parse sequence number
    if FTL_ATTR_BGP_ROUTE_SEQUENCE >= 0:
        route_record_prefix[FTL_ATTR_BGP_ROUTE_SEQUENCE] = struct_unpack(STRUCT_4B, entry_bytes[offset:offset + 4])[0]
    offset += 4

    # Parse AFI/SAFI-specific RIB entries
    if mrt_subtype in MRT_TABLE_DUMP_V2_RIB_SPECIFIC_ANY:

        # -----------------------------------------------
        # [RFC6396] 4.3.2. AFI/SAFI-Specific RIB Subtypes
        # -----------------------------------------------
        #  0                   1                   2                   3
        #  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # |                         Sequence Number                       |
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # | Prefix Length |
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # |                        Prefix (variable)                      |
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # |         Entry Count           |  RIB Entries (variable)
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

        try:
            # Parse prefix
            # NOTE: Human-readable conversions are done in unpack_mrt_bgp_nlri()
            preflen = (entry_bytes[offset] + 7) // 8 + 1
            afinet = AF_INET6 if mrt_subtype in MRT_TABLE_DUMP_V2_RIB_IPV6_ANY else AF_INET
            _, prefix_proto, prefix, _, _, _ = next(iter(unpack_mrt_bgp_nlri(stats_record, caches,
                                                                             entry_bytes[offset:offset + preflen],
                                                                             afinet)))
            offset += preflen

            # Add prefix protocol to route record prefix
            if FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL >= 0:
                route_record_prefix[FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL] = prefix_proto

            # Add prefix to route record prefix
            if FTL_ATTR_BGP_ROUTE_PREFIX >= 0:
                route_record_prefix[FTL_ATTR_BGP_ROUTE_PREFIX] = prefix

        # Yield (or re-raise) prefix errors
        except FtlMrtDataError as error:
            yield from route_error(error)
            return
        except StopIteration:
            yield from route_error(FtlMrtDataError('Unable to decode prefix for table dump v2 entry',
                                                   reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_PREFIX, data=entry_bytes))
            return

    # Parse generic RIB entries
    elif mrt_subtype in MRT_TABLE_DUMP_V2_RIB_GENERIC_ANY:

        # -------------------------------------
        # [RFC6396] 4.3.3.  RIB_GENERIC Subtype
        # -------------------------------------
        #  0                   1                   2                   3
        #  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # |                         Sequence Number                       |
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # |    Address Family Identifier  |Subsequent AFI |
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # |     Network Layer Reachability Information (variable)         |
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # |         Entry Count           |  RIB Entries (variable)
        # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

        # Parse AFI value
        afi = struct_unpack(STRUCT_2B, entry_bytes[offset:offset + 2])[0]
        offset += 2

        # Parse SAFI value
        safi = entry_bytes[offset]
        offset += 1

        # Parse prefix
        try:
            # Check AFI value
            if afi != AFI_IPV4 and afi != AFI_IPV6:  # pylint: disable=consider-using-in
                raise FtlMrtDataError(f'Invalid AFI value ({afi}) in table dump v2 RIB_GENERIC entry',
                                      reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_PROTOCOL, data=entry_bytes)

            # Check SAFI value
            if safi != BGP_SAFI_UNICAST and safi != BGP_SAFI_MULTICAST:  # pylint: disable=consider-using-in
                raise FtlMrtDataError(f'Unsupported SAFI value ({safi}) in table dump v2 RIB_GENERIC entry',
                                      reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_PROTOCOL, data=entry_bytes)

            # Parse prefix
            # NOTE: Human-readable conversions are done in unpack_mrt_bgp_nlri()
            preflen = (entry_bytes[offset] + 7) // 8 + 1
            afinet = AF_INET6 if afi == AFI_IPV6 else AF_INET
            _, prefix_proto, prefix, path_id, _, _ = next(iter(unpack_mrt_bgp_nlri(caches, stats_record,
                                                                                   entry_bytes[offset:offset + preflen],
                                                                                   afinet, addpath=addpath)))
            offset += preflen

            # Add prefix protocol to route record prefix
            if FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL >= 0:
                route_record_prefix[FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL] = prefix_proto

            # Add prefix to route record prefix
            if FTL_ATTR_BGP_ROUTE_PREFIX >= 0:
                route_record_prefix[FTL_ATTR_BGP_ROUTE_PREFIX] = prefix

            # Add ADD-PATH path ID to route record prefix
            addpath_rib = False
            if path_id is not None:
                if FTL_ATTR_BGP_ROUTE_PATH_ID >= 0:
                    route_record_prefix[FTL_ATTR_BGP_ROUTE_PATH_ID] = path_id

        # Yield (or re-raise) prefix errors
        except FtlMrtDataError as error:
            yield from route_error(error)
            return
        except StopIteration:
            yield from route_error(FtlMrtDataError('Unable to decode prefix for table dump v2 entry',
                                                   reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_PREFIX, data=entry_bytes))
            return

    # Parse unknown MRT subtype
    else:
        yield from route_error(FtlMrtHeaderError(f'Unknown MRT subtype ({mrt_subtype}) for table dump v2 entry',
                                                 reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_TYPE, data=entry_bytes))
        return

    # Parse entry count
    entry_count = struct_unpack(STRUCT_2B, entry_bytes[offset:offset + 2])[0]
    offset += 2

    # ----------------------------
    # [RFC6396] 4.3.4. RIB Entries
    # ----------------------------
    #  0                   1                   2                   3
    #  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |         Peer Index            |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                         Originated Time                       |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |      Attribute Length         |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                    BGP Attributes... (variable)
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    # Sanitize RIB entries
    static_offset = 8 if addpath is False else 12
    cur_entries, cur_offset, end_offset = 0, offset + static_offset, len(entry_bytes) + static_offset
    while cur_offset < end_offset:
        cur_offset += static_offset + struct_unpack(STRUCT_2B, entry_bytes[cur_offset - 2:cur_offset])[0]
        cur_entries += 1

    # ---------------------------------------------
    # [RFC8050] 4.1. AFI/SAFI-Specific RIB Subtypes
    # ---------------------------------------------
    #  0                   1                   2                   3
    #  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |         Peer Index            |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                         Originated Time                       |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                         Path Identifier                       |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |      Attribute Length         |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                    BGP Attributes... (variable)
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    # Handle invalid RIB entries
    if cur_entries != entry_count or cur_offset != end_offset:

        # Prepare data error
        if addpath_error is None:
            addpath_error = FtlMrtDataError('Unable to decode table dump v2 RIB entries',
                                            reason=FTL_ATTR_BGP_ERROR_REASON_MISSING_DATA, data=entry_bytes)

        # Retry with ADD-PATH support
        # NOTE: Some MRT exporters include path IDs but fail to set an ADD-PATH entry subtype (RFC7911)
        if addpath is False:
            yield from unpack_mrt_entry_tdv2_rib(caches, stats_record, route_records, entry_bytes, mrt_subtype,
                                                 peer_table, addpath=True, addpath_error=addpath_error)
            return

        # Retry failed
        yield from route_error(addpath_error)
        return

    # Check for ADD-PATH fixes
    if addpath is True and addpath_error is not None:

        # Update stats record
        if FTL_RECORD_BGP_STATS:

            # Add ADD-PATH fix
            if FTL_ATTR_BGP_STATS_MRT_FIXES >= 0:
                fixtype = str(FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_ADDPATH)
                if FTL_ATTR_BGP_STATS_MRT_FIXES_HUMAN:
                    fixtype = FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_ADDPATH_STR
                stats_record_fix = stats_record[FTL_ATTR_BGP_STATS_MRT_FIXES]
                stats_record_fix[fixtype] = stats_record_fix.get(fixtype, 0) + 1

    # Parse RIB entries
    for _ in range(entry_count):

        # Initialize route record
        route_record = list(route_record_prefix)

        # Parse peer index and access peer table
        # NOTE: Human-readable conversions are done in unpack_mrt_entry_tdv2_index()
        if (FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL >= 0 or FTL_ATTR_BGP_ROUTE_PEER_BGP_ID >= 0
            or FTL_ATTR_BGP_ROUTE_PEER_AS >= 0 or FTL_ATTR_BGP_ROUTE_PEER_IP >= 0):
            peer_index = struct_unpack(STRUCT_2B, entry_bytes[offset:offset + 2])[0]
            peer_proto, peer_bgp_id, peer_as, peer_ip = peer_table[peer_index]
            if FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL >= 0:
                route_record[FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL] = peer_proto
            if FTL_ATTR_BGP_ROUTE_PEER_BGP_ID >= 0:
                route_record[FTL_ATTR_BGP_ROUTE_PEER_BGP_ID] = peer_bgp_id
            if FTL_ATTR_BGP_ROUTE_PEER_AS >= 0:
                route_record[FTL_ATTR_BGP_ROUTE_PEER_AS] = peer_as
            if FTL_ATTR_BGP_ROUTE_PEER_IP >= 0:
                route_record[FTL_ATTR_BGP_ROUTE_PEER_IP] = peer_ip
        offset += 2

        # Parse timestamp
        if FTL_ATTR_BGP_ROUTE_TIMESTAMP >= 0:
            ts = float(struct_unpack(STRUCT_4B, entry_bytes[offset:offset + 4])[0])
            if FTL_ATTR_BGP_ROUTE_TIMESTAMP_HUMAN:

                # Cache date string for minute-based timestamp
                if caches:
                    ts_cache = caches[CACHE_TS]
                    ts_min, ts_sec = divmod(ts, 60)
                    ts_cached = ts_cache.get(ts_min, None)
                    if ts_cached is None:
                        ts_cached = datetime_utcfromtimestamp(ts).strftime(DATETIME_FORMAT_MIN)
                        ts_cache[ts_min] = ts_cached

                    # Add seconds and microseconds
                    ts = f'{ts_cached}:{ts_sec:09.6f}'

                # Do not use cache
                else:
                    ts = datetime_utcfromtimestamp(ts).strftime(DATETIME_FORMAT_USEC)

            # Add timestamp
            route_record[FTL_ATTR_BGP_ROUTE_TIMESTAMP] = ts
        offset += 4

        # Parse ADD-PATH path ID
        if addpath_rib is True:
            route_record[FTL_ATTR_BGP_ROUTE_PATH_ID] = struct_unpack(STRUCT_4B, entry_bytes[offset:offset + 4])[0]
            offset += 4

        # Parse attributes length
        alen = struct_unpack(STRUCT_2B, entry_bytes[offset:offset + 2])[0]
        offset += 2

        # Parse attributes
        attr_bytes = entry_bytes[offset:offset + alen]
        offset += alen
        try:
            unpack_mrt_bgp_attr(caches, stats_record, route_init, route_emit, route_record, attr_bytes, aslen=4,
                                addpath=addpath, rib=True)

        # Yield (or re-raise) attribute errors
        except FtlMrtDataError as error:
            yield from route_error(error)
            continue

        # Update stats record
        if FTL_RECORD_BGP_STATS:

            # Update IPv4 RIB routes
            if afinet == AF_INET:
                if FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV4 >= 0:
                    stats_record[FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV4] += 1

            # Update IPv6 RIB routes
            if afinet == AF_INET6:
                if FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV6 >= 0:
                    stats_record[FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV6] += 1

        # Yield final route record
        if FTL_RECORD_BGP_ROUTE:
            yield route_emit(route_record)



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# Local imports
from .const import MRT_BGP_UPDATE
from .const import MRT_BGP_OPEN
from .const import MRT_BGP_NOTIFY
from .const import MRT_BGP_KEEPALIVE
from .const import MRT_BGP_STATE_CHANGE
from ..bgp.const import BGP_BGP4MP_UPDATE
from ..bgp.const import BGP_BGP4MP_KEEPALIVE
from ..bgp.const import BGP_BGP4MP_NOTIFICATION
from ..bgp.const import BGP_BGP4MP_OPEN
from ..bgp.msg import unpack_mrt_bgp_msg
from ...util import CACHE_TS
from ...const import IPV4
from ...const import IPV4_STR
from ...const import AF_INET
from ...const import STRUCT_2B
from ...const import STRUCT_4B
from ...const import DATETIME_FORMAT_USEC
from ...const import DATETIME_FORMAT_MIN
from ...const import struct_unpack
from ...const import socket_inet_ntop
from ...const import datetime_utcfromtimestamp
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_TIMESTAMP
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_PEER_PROTOCOL
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_PEER_AS
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_PEER_IP
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_OLD_STATE
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_NEW_STATE
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_TIMESTAMP_HUMAN
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_PEER_PROTOCOL_HUMAN
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_PEER_IP_HUMAN
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_OLD_STATE_HUMAN
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_NEW_STATE_HUMAN
from ....model.const import FTL_ATTR_BGP_STATE_CHANGE_STATE_TO_STR
from ....model.record import FTL_RECORD_BGP_STATE_CHANGE


def unpack_mrt_entry_bgp(caches, stats_record, bgp_error, state_change_records, route_records, keep_alive_records,
                         route_refresh_records, notification_records, open_records, entry_bytes, mrt_subtype, sequence,
                         ts):
    """ Parse MRT BGP entry.
    """
    ###############
    # BGP PARSING #
    ###############

    # Prepare byte offset
    offset = 0

    ####################
    # BGP STATE_CHANGE #
    ####################

    # -------------------------------------------
    # [draft-ietf-grow-mrt-07] BGP_UPDATE Subtype
    # -------------------------------------------
    #  0                   1                   2                   3
    #  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |         Peer AS number        |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                        Peer IP address                        |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |            Old State          |          New State            |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    # Parse state change message
    if mrt_subtype == MRT_BGP_STATE_CHANGE:

        # Access state change template
        state_change_init, state_change_emit, _ = state_change_records

        # Initialize state change record
        state_change_record = list(state_change_init)

        # Add timestamp to state change record
        if FTL_ATTR_BGP_STATE_CHANGE_TIMESTAMP >= 0:
            ts_state_change = ts

            # Cache date string for minute-based timestamp
            if FTL_ATTR_BGP_STATE_CHANGE_TIMESTAMP_HUMAN:
                if caches:
                    ts_cache = caches[CACHE_TS]
                    ts_min, ts_sec = divmod(ts_state_change, 60)
                    ts_cached = ts_cache.get(ts_min, None)
                    if ts_cached is None:
                        ts_cached = datetime_utcfromtimestamp(ts).strftime(DATETIME_FORMAT_MIN)
                        ts_cache[ts_min] = ts_cached

                    # Add seconds and microseconds
                    ts_state_change = f'{ts_cached}:{ts_sec:09.6f}'

                # Do not use cache
                else:
                    ts_state_change = datetime_utcfromtimestamp(ts_state_change).strftime(DATETIME_FORMAT_USEC)

            # Add timestamp
            state_change_record[FTL_ATTR_BGP_STATE_CHANGE_TIMESTAMP] = ts_state_change

        # Add peer protocol to state change record
        if FTL_ATTR_BGP_STATE_CHANGE_PEER_PROTOCOL >= 0:
            if FTL_ATTR_BGP_STATE_CHANGE_PEER_PROTOCOL_HUMAN:
                state_change_record[FTL_ATTR_BGP_STATE_CHANGE_PEER_PROTOCOL] = IPV4_STR
            else:
                state_change_record[FTL_ATTR_BGP_STATE_CHANGE_PEER_PROTOCOL] = IPV4

        # Add peer AS to state change record
        if FTL_ATTR_BGP_STATE_CHANGE_PEER_AS >= 0:
            peer_as = struct_unpack(STRUCT_2B, entry_bytes[offset:offset + 2])[0]
            state_change_record[FTL_ATTR_BGP_STATE_CHANGE_PEER_AS] = peer_as
        offset += 2

        # Add peer IP to state change record
        if FTL_ATTR_BGP_STATE_CHANGE_PEER_IP >= 0:
            peer_ip = entry_bytes[offset:offset + 4]
            if FTL_ATTR_BGP_STATE_CHANGE_PEER_IP_HUMAN:
                peer_ip = socket_inet_ntop(AF_INET, peer_ip)
            else:
                peer_ip = struct_unpack(STRUCT_4B, peer_ip)[0]
            state_change_record[FTL_ATTR_BGP_STATE_CHANGE_PEER_IP] = peer_ip
        offset += 4

        # Add old state to state change record
        if FTL_ATTR_BGP_STATE_CHANGE_OLD_STATE >= 0:
            old_state = struct_unpack(STRUCT_2B, entry_bytes[offset:offset + 2])[0]
            if FTL_ATTR_BGP_STATE_CHANGE_OLD_STATE_HUMAN:
                old_state = FTL_ATTR_BGP_STATE_CHANGE_STATE_TO_STR.get(old_state, str(old_state))
            state_change_record[FTL_ATTR_BGP_STATE_CHANGE_OLD_STATE] = old_state
        offset += 2

        # Add new state to state change record
        if FTL_ATTR_BGP_STATE_CHANGE_NEW_STATE >= 0:
            new_state = struct_unpack(STRUCT_2B, entry_bytes[offset:offset + 2])[0]
            if FTL_ATTR_BGP_STATE_CHANGE_NEW_STATE_HUMAN:
                new_state = FTL_ATTR_BGP_STATE_CHANGE_STATE_TO_STR.get(new_state, str(new_state))
            state_change_record[FTL_ATTR_BGP_STATE_CHANGE_NEW_STATE] = new_state
        offset += 2

        # Yield final state change record
        if FTL_RECORD_BGP_STATE_CHANGE:
            yield state_change_emit(state_change_record)
        return

    ##############
    # BGP UPDATE #
    ##############

    # -------------------------------------------
    # [draft-ietf-grow-mrt-07] BGP_UPDATE Subtype
    # -------------------------------------------
    #  0                   1                   2                   3
    #  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |         Peer AS number        |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                         Peer IP address                       |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |        Local AS number        |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                        Local IP address                       |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                    BGP UPDATE Contents (variable)
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    # Parse peer AS
    peer_as = struct_unpack(STRUCT_2B, entry_bytes[offset:offset + 2])[0]
    offset += 2

    # Parse peer IP
    peer_ip = entry_bytes[offset:offset + 4]
    offset += 4

    # Skip local AS
    offset += 2

    # Skip local IP
    offset += 4

    # Parse message type
    mtype = None
    if mrt_subtype == MRT_BGP_UPDATE:
        mtype = BGP_BGP4MP_UPDATE
    elif mrt_subtype == MRT_BGP_KEEPALIVE:
        mtype = BGP_BGP4MP_KEEPALIVE
    elif mrt_subtype == MRT_BGP_NOTIFY:
        mtype = BGP_BGP4MP_NOTIFICATION
    elif mrt_subtype == MRT_BGP_OPEN:
        mtype = BGP_BGP4MP_OPEN

    # Parse message
    if mtype is not None:
        msg_bytes = entry_bytes[offset:]
        yield from unpack_mrt_bgp_msg(caches, stats_record, bgp_error, route_records, keep_alive_records,
                                      route_refresh_records, notification_records, open_records, msg_bytes, mtype,
                                      sequence, ts, peer_as, peer_ip, AF_INET, aslen=2)



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# Local imports
from ...const import IPV4
from ...const import IPV6
from ...const import IPV4_STR
from ...const import IPV6_STR
from ...const import AF_INET
from ...const import AF_INET6
from ...const import STRUCT_4B
from ...const import STRUCT_8B8B
from ...const import struct_unpack
from ...const import socket_inet_ntop
from ....model.attr import FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL
from ....model.attr import FTL_ATTR_BGP_ROUTE_PREFIX
from ....model.attr import FTL_ATTR_BGP_ROUTE_PATH_ID
from ....model.attr import FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_PREFIX_HUMAN
from ....model.attr import FTL_ATTR_BGP_STATS_MRT_FIXES
from ....model.attr import FTL_ATTR_BGP_STATS_MRT_FIXES_HUMAN
from ....model.const import FTL_ATTR_BGP_STATS_MRT_FIXES_BGP4MP_ADDPATH
from ....model.const import FTL_ATTR_BGP_STATS_MRT_FIXES_BGP4MP_ADDPATH_STR
from ....model.const import FTL_ATTR_BGP_ERROR_REASON_INVALID_PREFIX
from ....model.record import FTL_RECORD_BGP_STATS
from ....model.error import FtlMrtDataError


def unpack_mrt_bgp_nlri(caches, stats_record, nlri_bytes, afinet, nexthop_proto=None, nexthop_ip=None, addpath=False,
                        addpath_error=None, relaxed=False):
    """ Parse MRT NLRI list.
    """
    # Initialize NLRI list and protocol
    nlri, prefix_proto = list(), None

    # Parse protocol
    iplen, ipbits = (16, 128) if afinet == AF_INET6 else (4, 32)
    if FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL >= 0:
        if FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL_HUMAN:
            prefix_proto = IPV6_STR if afinet == AF_INET6 else IPV4_STR
        else:
            prefix_proto = IPV6 if afinet == AF_INET6 else IPV4

    # --------------------------
    # [RFC4760] 5. NLRI Encoding
    # --------------------------
    # The Network Layer Reachability information is encoded as one or more
    # 2-tuples of the form <length, prefix>, whose fields are described
    # below:
    #
    # +---------------------------+
    # |   Length (1 octet)        |
    # +---------------------------+
    # |   Prefix (variable)       |
    # +---------------------------+
    #
    # The use and the meaning of these fields are as follows:
    #
    # a) Length:
    #
    #    The Length field indicates the length, in bits, of the address
    #    prefix. A length of zero indicates a prefix that matches all (as
    #    specified by the address family) addresses (with prefix, itself,
    #    of zero octets).
    #
    # b) Prefix:
    #
    #    The Prefix field contains an address prefix followed by enough
    #    trailing bits to make the end of the field fall on an octet
    #    boundary. Note that the value of trailing bits is irrelevant.

    # Prepare byte offset
    offset = 0

    # Parse NLRI list
    try:
        nlrilen = len(nlri_bytes)
        while offset < nlrilen:

            # Initialize prefix and ADD-PATH path ID
            prefix, path_id = None, None

            # ------------------------------------
            # [RFC7911] 3. Extended NLRI Encodings
            # ------------------------------------
            # In order to carry the Path Identifier in an UPDATE message, the NLRI
            # encoding MUST be extended by prepending the Path Identifier field,
            # which is of four octets.
            #
            # For example, the NLRI encoding specified in [RFC4271] is extended as
            # the following:
            #
            # +--------------------------------+
            # | Path Identifier (4 octets)     |
            # +--------------------------------+
            # | Length (1 octet)               |
            # +--------------------------------+
            # | Prefix (variable)              |
            # +--------------------------------+
            #
            # The Path Identifier specified in Section 3 can be used to advertise
            # multiple paths for the same address prefix without subsequent
            # advertisements replacing the previous ones. Apart from the fact that
            # this is now possible, the route advertisement rules of [RFC4271] are
            # not changed. In particular, a new advertisement for a given address
            # prefix and a given Path Identifier replaces a previous advertisement
            # for the same address prefix and Path Identifier. If a BGP speaker
            # receives a message to withdraw a prefix with a Path Identifier not
            # seen before, it SHOULD silently ignore it.
            #
            # A BGP speaker SHOULD include the best route [RFC4271] when more than
            # one path is advertised to a neighbor, unless it is a path received
            # from that neighbor.
            #
            # As the Path Identifiers are locally assigned, and may or may not be
            # persistent across a control plane restart of a BGP speaker.

            # Handle ADD-PATH path ID
            if addpath is True:

                # Check remaining bytes
                if offset + 5 > nlrilen:
                    raise ValueError('incomplete ADD-PATH prefix')

                # Parse ADD-PATH path ID
                if FTL_ATTR_BGP_ROUTE_PATH_ID >= 0:
                    path_id = struct_unpack(STRUCT_4B, nlri_bytes[offset:offset + 4])[0]
                offset += 4

            # Parse prefix mask
            mask = nlri_bytes[offset]
            plen = (mask + 7) // 8
            offset += 1

            # Check remaining bytes
            if offset + plen > nlrilen:
                raise ValueError('incomplete prefix')

            # Parse prefix bytes
            prefix_bytes = nlri_bytes[offset:offset + plen]
            if plen < iplen:
                prefix_bytes = b''.join((prefix_bytes, bytearray(iplen - plen)))
            offset += plen

            # Check for invalid mask
            if mask > ipbits:
                raise ValueError(f'invalid {IPV6_STR if afinet == AF_INET6 else IPV4_STR} mask (/{mask})')

            # Parse prefix address
            prefix_int = 0
            if afinet == AF_INET6:
                net, host = struct_unpack(STRUCT_8B8B, prefix_bytes)
                prefix_int = (net << 64) + host
            else:
                prefix_int = struct_unpack(STRUCT_4B, prefix_bytes)[0]

            # Check for invalid address (passes #1 and #2 only)
            if relaxed is False:

                # Blacklist default routes
                if mask == 0:
                    raise ValueError(f'invalid {IPV6_STR if afinet == AF_INET6 else IPV4_STR} default route')

                # Blacklist 0.0.0.0/8 and 240.0.0.0/4 [RFC791/RFC1112]
                if afinet == AF_INET:
                    if not 0xffffff < prefix_int < 0xf0000000:
                        raise ValueError(f'invalid {IPV4_STR} address ({socket_inet_ntop(afinet, prefix_bytes)}/{mask})')

                # Whitelist 2000::/3 [RFC4291]
                elif mask < 3 or prefix_int >> 125 != 1:  # pylint: disable=confusing-consecutive-elif

                    # Whitelist fc00::/7 [RFC8190]
                    if mask < 7 or prefix_int >> 120 not in {0xfc, 0xfd}:

                        # Whitelist ff00::/8 [RFC2373]
                        if mask < 8 or prefix_int >> 120 != 0xff:

                            # Whitelist fe80::/10 [RFC4291]
                            if mask < 10 or prefix_int >> 118 != 0x3fa:

                                # Whitelist 2620:4f:8000::/48 and 64:ff9b:1::/48 [RFC7534/RFC8215]
                                if mask < 48 or prefix_int >> 80 not in {0x2620004f8000, 0x64ff9b0001}:

                                    # Whitelist 100::/64 [RFC6666]
                                    if mask < 64 or prefix_int >> 64 != 0x100000000000000:

                                        # Whitelist 64:ff9b::/96 and ::ffff:0:0/96 (and 0::/96) [RFC6052/RFC4291]
                                        if mask < 96 or prefix_int >> 32 not in {0x64ff9b0000000000000000, 0xffff, 0x0}:

                                            # Invalid ipv6 prefix
                                            raise ValueError(f'invalid {IPV6_STR} address '
                                                             f'({socket_inet_ntop(afinet, prefix_bytes)}/{mask})')

            # Check for invalid alignment
            shift = ipbits - mask
            if (prefix_int >> shift) << shift != prefix_int:
                raise ValueError(f'misaligned prefix ({socket_inet_ntop(afinet, prefix_bytes)}/{mask})')

            # Parse prefix
            if FTL_ATTR_BGP_ROUTE_PREFIX >= 0:
                if FTL_ATTR_BGP_ROUTE_PREFIX_HUMAN:
                    prefix = f'{socket_inet_ntop(afinet, prefix_bytes)}/{mask}'
                else:
                    prefix = (prefix_int, mask)

            # Update NLRI list
            nlri.append((afinet, prefix_proto, prefix, path_id, nexthop_proto, nexthop_ip))

    # Handle prefix errors
    except ValueError as exc:

        # Prepare data error
        first_try = False
        if addpath_error is None:
            first_try = True
            addpath_error = FtlMrtDataError('Unable to decode NLRI entries',
                                            reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_PREFIX, data=nlri_bytes,
                                            exception=exc)

        # Allow for less strict NLRI parsing
        if relaxed is False:

            # Retry with ADD-PATH support
            # NOTE: Some MRT exporters include path IDs but fail to set an ADD-PATH message subtype (RFC7911)
            if addpath is False:
                return unpack_mrt_bgp_nlri(caches, stats_record, nlri_bytes, afinet, nexthop_proto=nexthop_proto,
                                           nexthop_ip=nexthop_ip, addpath=True, addpath_error=addpath_error)

            # Retry with relaxed prefix validation (after ADD-PATH retry failed)
            if first_try is False:
                return unpack_mrt_bgp_nlri(caches, stats_record, nlri_bytes, afinet, nexthop_proto=nexthop_proto,
                                           nexthop_ip=nexthop_ip, addpath=False, addpath_error=addpath_error,
                                           relaxed=True)

        # All retries failed
        raise addpath_error  # pylint: disable=raise-missing-from

    # Check for ADD-PATH fixes
    if addpath is True and addpath_error is not None:

        # Update stats record
        if FTL_RECORD_BGP_STATS:

            # Add ADD-PATH fix
            if FTL_ATTR_BGP_STATS_MRT_FIXES >= 0:
                fixtype = str(FTL_ATTR_BGP_STATS_MRT_FIXES_BGP4MP_ADDPATH)
                if FTL_ATTR_BGP_STATS_MRT_FIXES_HUMAN:
                    fixtype = FTL_ATTR_BGP_STATS_MRT_FIXES_BGP4MP_ADDPATH_STR
                stats_record_fix = stats_record[FTL_ATTR_BGP_STATS_MRT_FIXES]
                stats_record_fix[fixtype] = stats_record_fix.get(fixtype, 0) + 1

    # Return NRLI listTTR_BGP_STATS_MRT_FIXES_BGP4MP_ADDPATH
    return tuple(nlri)



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# System imports
import re

# Routing table formats
TABLE_REGEX_STD = re.compile(r'^(\s+)(Network\s+)(Next Hop\s*)(\sMetric)(\s+LocPrf)(\s+Weight)(\s+)Path\s*$')
TABLE_REGEX_ALT = re.compile(r'^(\s+P)(\s+Pref)(\s+)(Time\s+)(Destination\s+)(Next Hop\s+)(If\s+)Path\s*$')


def check_lgl_table(header):
    """ Match looking glass header line against supported table formats.
    """
    # Check for standard table header (PCH)
    match = TABLE_REGEX_STD.match(header)
    if match is not None:
        return unpack_lgl_entry_std(*(len(g) for g in match.groups()))

    # Check for alternative table header (early RIPE RIS)
    match = TABLE_REGEX_ALT.match(header)
    if match is not None:
        return unpack_lgl_entry_alt(*(len(g) for g in match.groups()))

    # Table header not supported
    return None


def unpack_lgl_entry_std(status_skip, prefix_len, nexthop_len, med_len, locpref_len, weight_len, path_skip):
    """ Parse looking glass entry in standard table format (PCH).
    """
    # ------------------------
    # [PCH] % sh bgp ipv4 wide
    # ------------------------
    #    Network          Next Hop            Metric LocPrf Weight Path
    # *  1.0.0.0/24       206.126.236.19           0             0 3257 13335 i
    # *>                  206.126.237.30           0             0 13335 i
    # *> 1.0.16.0/24      206.126.236.23           0             0 2497 2519 i

    # Prepare table indexes
    prefix_start = status_skip
    prefix_end = prefix_start + prefix_len
    nexthop_start = prefix_end
    nexthop_end = nexthop_start + nexthop_len
    med_start = nexthop_end
    med_end = med_start + med_len
    locpref_start = med_end
    locpref_end = locpref_start + locpref_len
    aspath_start = locpref_end + weight_len + path_skip
    entry_len_min = aspath_start

    def unpack_lgl_entry(entry):
        """ Collector-specific entry parser.
        """
        # Check entry
        if len(entry) <= entry_len_min:
            return None

        # Parse prefix
        prefix = entry[prefix_start:prefix_end].rstrip()
        if len(prefix) == prefix_end - prefix_start:
            prefix, entry = entry[prefix_start:].split(None, 1)
            entry = '_' * prefix_end + entry.lstrip()
            if len(entry) <= entry_len_min:
                return None

        # Parse nexthop
        nexthop = entry[nexthop_start:nexthop_end].rstrip()
        if len(nexthop) == nexthop_end - nexthop_start:
            nexthop, entry = entry[nexthop_start:].split(None, 1)
            med, _ = entry.split(None, 1)
            entry = '.' * nexthop_end + ' ' * (med_len - len(med)) + entry
            # NOTE: If there is no MED metric, we are unable to parse 3-line entries
            if len(entry) <= entry_len_min:
                return None

        # Parse MED metric and local preference
        med = entry[med_start:med_end].strip()
        locpref = entry[locpref_start:locpref_end].lstrip()

        # Parse AS path
        aspath, origin = '', entry[aspath_start:].rstrip()
        if len(origin) > 1:
            aspath, origin = origin.rsplit(None, 1)

        # Return parsed entry
        return prefix, nexthop, med, locpref, aspath, origin

    # Return unpack function
    return unpack_lgl_entry


def unpack_lgl_entry_alt(status_skip, locpref_len, time_skip, time_len, prefix_len, nexthop_len, iface_len):
    """ Parse looking glass entry in alternative table format (early RIPE RIS).
    """
    # -------------------------
    # [RIPE] % sh bgp ipv4 wide
    # -------------------------
    #   P Pref Time     Destination                Next Hop                 If      Path
    # > B    0 11:53:59 3.0.0.0/8                  193.0.0.56               eth0    3333 286 701 80 i
    # * B    0 11:53:46 3.0.0.0/8                  193.0.0.59               eth0    3333 286 701 80 i

    # Prepare table indexes
    locpref_start = status_skip
    locpref_end = locpref_start + locpref_len
    prefix_start = locpref_end + time_skip + time_len
    prefix_end = prefix_start + prefix_len
    nexthop_start = prefix_end
    nexthop_end = nexthop_start + nexthop_len
    aspath_start = nexthop_end + iface_len
    entry_len_min = aspath_start

    def unpack_lgl_entry(entry):
        """ Collector-specific entry parser.
        """
        # Check entry
        if len(entry) <= entry_len_min:
            return None

        # Parse local preference
        locpref = entry[locpref_start:locpref_end].lstrip()

        # Parse prefix
        prefix = entry[prefix_start:prefix_end].rstrip()
        if len(prefix) == prefix_end - prefix_start:
            prefix, entry = entry[prefix_start:].split(None, 1)
            entry = '_' * prefix_end + entry.lstrip()
            if len(entry) <= entry_len_min:
                return None

        # Parse nexthop
        nexthop = entry[nexthop_start:nexthop_end].rstrip()
        if len(nexthop) == nexthop_end - nexthop_start:
            nexthop, entry = entry[nexthop_start:].split(None, 1)
            entry = '.' * nexthop_end + entry.lstrip()

        # Parse AS path
        aspath, origin = '', entry[aspath_start:].rstrip()
        if len(origin) > 1:
            aspath, origin = origin.rsplit(None, 1)

        # Return parsed entry
        return prefix, nexthop, '', locpref, aspath, origin

    # Return unpack function
    return unpack_lgl_entry



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
# flake8: noqa
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# Local imports
from ..data.const import AFI_IPV4
from ..data.const import AFI_IPV6
from ..data.mrt.entry.const import MRT_NULL
from ..data.mrt.entry.const import MRT_START
from ..data.mrt.entry.const import MRT_DIE
from ..data.mrt.entry.const import MRT_I_AM_DEAD
from ..data.mrt.entry.const import MRT_PEER_DOWN
from ..data.mrt.entry.const import MRT_BGP
from ..data.mrt.entry.const import MRT_RIP
from ..data.mrt.entry.const import MRT_IDRP
from ..data.mrt.entry.const import MRT_RIPNG
from ..data.mrt.entry.const import MRT_BGP4PLUS
from ..data.mrt.entry.const import MRT_BGP4PLUS_01
from ..data.mrt.entry.const import MRT_OSPFV2
from ..data.mrt.entry.const import MRT_TABLE_DUMP
from ..data.mrt.entry.const import MRT_TABLE_DUMP_V2
from ..data.mrt.entry.const import MRT_BGP4MP
from ..data.mrt.entry.const import MRT_BGP4MP_ET
from ..data.mrt.entry.const import MRT_ISIS
from ..data.mrt.entry.const import MRT_ISIS_ET
from ..data.mrt.entry.const import MRT_OSPFV3
from ..data.mrt.entry.const import MRT_OSPFV3_ET
from ..data.mrt.entry.const import MRT_BGP_NULL
from ..data.mrt.entry.const import MRT_BGP_UPDATE
from ..data.mrt.entry.const import MRT_BGP_PREF_UPDATE
from ..data.mrt.entry.const import MRT_BGP_STATE_CHANGE
from ..data.mrt.entry.const import MRT_BGP_SYNC
from ..data.mrt.entry.const import MRT_BGP_OPEN
from ..data.mrt.entry.const import MRT_BGP_NOTIFY
from ..data.mrt.entry.const import MRT_BGP_KEEPALIVE
from ..data.mrt.entry.const import MRT_BGP4MP_ENTRY
from ..data.mrt.entry.const import MRT_BGP4MP_SNAPSHOT
from ..data.mrt.entry.const import MRT_BGP4MP_STATE_CHANGE
from ..data.mrt.entry.const import MRT_BGP4MP_STATE_CHANGE_AS4
from ..data.mrt.entry.const import MRT_BGP4MP_MESSAGE
from ..data.mrt.entry.const import MRT_BGP4MP_MESSAGE_ADDPATH
from ..data.mrt.entry.const import MRT_BGP4MP_MESSAGE_LOCAL
from ..data.mrt.entry.const import MRT_BGP4MP_MESSAGE_LOCAL_ADDPATH
from ..data.mrt.entry.const import MRT_BGP4MP_MESSAGE_AS4
from ..data.mrt.entry.const import MRT_BGP4MP_MESSAGE_AS4_ADDPATH
from ..data.mrt.entry.const import MRT_BGP4MP_MESSAGE_AS4_LOCAL
from ..data.mrt.entry.const import MRT_BGP4MP_MESSAGE_AS4_LOCAL_ADDPATH
from ..data.mrt.entry.const import MRT_TABLE_DUMP_V2_PEER_INDEX_TABLE
from ..data.mrt.entry.const import MRT_TABLE_DUMP_V2_GEO_PEER_TABLE
from ..data.mrt.entry.const import MRT_TABLE_DUMP_V2_RIB_GENERIC
from ..data.mrt.entry.const import MRT_TABLE_DUMP_V2_RIB_GENERIC_ADDPATH
from ..data.mrt.entry.const import MRT_TABLE_DUMP_V2_RIB_IPV4_UNICAST
from ..data.mrt.entry.const import MRT_TABLE_DUMP_V2_RIB_IPV4_MULTICAST
from ..data.mrt.entry.const import MRT_TABLE_DUMP_V2_RIB_IPV6_UNICAST
from ..data.mrt.entry.const import MRT_TABLE_DUMP_V2_RIB_IPV6_MULTICAST
from ..data.mrt.entry.const import MRT_TABLE_DUMP_V2_RIB_IPV4_UNICAST_ADDPATH
from ..data.mrt.entry.const import MRT_TABLE_DUMP_V2_RIB_IPV4_MULTICAST_ADDPATH
from ..data.mrt.entry.const import MRT_TABLE_DUMP_V2_RIB_IPV6_UNICAST_ADDPATH
from ..data.mrt.entry.const import MRT_TABLE_DUMP_V2_RIB_IPV6_MULTICAST_ADDPATH
from ..data.mrt.bgp.const import BGP_BGP4MP_RESERVED
from ..data.mrt.bgp.const import BGP_BGP4MP_OPEN
from ..data.mrt.bgp.const import BGP_BGP4MP_UPDATE
from ..data.mrt.bgp.const import BGP_BGP4MP_NOTIFICATION
from ..data.mrt.bgp.const import BGP_BGP4MP_KEEPALIVE
from ..data.mrt.bgp.const import BGP_BGP4MP_ROUTE_REFRESH
from ..data.mrt.bgp.const import BGP_STATE_IDLE
from ..data.mrt.bgp.const import BGP_STATE_CONNECT
from ..data.mrt.bgp.const import BGP_STATE_ACTIVE
from ..data.mrt.bgp.const import BGP_STATE_OPEN_SENT
from ..data.mrt.bgp.const import BGP_STATE_OPEN_CONFIRM
from ..data.mrt.bgp.const import BGP_STATE_ESTABLISHED
from ..data.mrt.bgp.const import BGP_STATE_CLEARING
from ..data.mrt.bgp.const import BGP_STATE_DELETED
from ..data.mrt.bgp.const import BGP_CAPABILITY_RESERVED_0
from ..data.mrt.bgp.const import BGP_CAPABILITY_BGP4MP
from ..data.mrt.bgp.const import BGP_CAPABILITY_ROUTE_REFRESH
from ..data.mrt.bgp.const import BGP_CAPABILITY_OUTBOUND_FILTER
from ..data.mrt.bgp.const import BGP_CAPABILITY_MULTIPLE_ROUTES
from ..data.mrt.bgp.const import BGP_CAPABILITY_EXTENDED_NEXT_HOP
from ..data.mrt.bgp.const import BGP_CAPABILITY_BGP4MP_ET
from ..data.mrt.bgp.const import BGP_CAPABILITY_BGPSEC
from ..data.mrt.bgp.const import BGP_CAPABILITY_MULTIPLE_LABELS
from ..data.mrt.bgp.const import BGP_CAPABILITY_BGP_ROLE
from ..data.mrt.bgp.const import BGP_CAPABILITY_GRACEFUL_RESTART
from ..data.mrt.bgp.const import BGP_CAPABILITY_AS4
from ..data.mrt.bgp.const import BGP_CAPABILITY_UNKNOWN_66
from ..data.mrt.bgp.const import BGP_CAPABILITY_DYNAMIC
from ..data.mrt.bgp.const import BGP_CAPABILITY_MULTISESSION
from ..data.mrt.bgp.const import BGP_CAPABILITY_ADDPATH
from ..data.mrt.bgp.const import BGP_CAPABILITY_ENHANCED_ROUTE_REFRESH
from ..data.mrt.bgp.const import BGP_CAPABILITY_LLGR
from ..data.mrt.bgp.const import BGP_CAPABILITY_POLICY_DISTRIBUTION
from ..data.mrt.bgp.const import BGP_CAPABILITY_FQDN
from ..data.mrt.bgp.const import BGP_CAPABILITY_BFD
from ..data.mrt.bgp.const import BGP_CAPABILITY_SOFTWARE_VERSION
from ..data.mrt.bgp.const import BGP_CAPABILITY_PRESTD_ROUTE_REFRESH
from ..data.mrt.bgp.const import BGP_CAPABILITY_PRESTD_POLICY_DISTRIBUTION
from ..data.mrt.bgp.const import BGP_CAPABILITY_PRESTD_OUTBOUND_FILTER
from ..data.mrt.bgp.const import BGP_CAPABILITY_PRESTD_MULTISESSION
from ..data.mrt.bgp.const import BGP_CAPABILITY_PRESTD_FQDN
from ..data.mrt.bgp.const import BGP_CAPABILITY_PRESTD_OPERATIONAL
from ..data.mrt.bgp.const import BGP_CAPABILITY_RESERVED_255
from ..data.mrt.bgp.const import BGP_ERROR_RESERVED
from ..data.mrt.bgp.const import BGP_ERROR_MESSAGE_HEADER
from ..data.mrt.bgp.const import BGP_ERROR_OPEN_MESSAGE
from ..data.mrt.bgp.const import BGP_ERROR_UPDATE_MESSAGE
from ..data.mrt.bgp.const import BGP_ERROR_HOLD_TIMER_EXPIRED
from ..data.mrt.bgp.const import BGP_ERROR_FSM
from ..data.mrt.bgp.const import BGP_ERROR_CEASE
from ..data.mrt.bgp.const import BGP_ERROR_ROUTE_REFRESH
from ..data.mrt.bgp.const import BGP_ERROR_MESSAGE_HEADER_UNSPECIFIC
from ..data.mrt.bgp.const import BGP_ERROR_MESSAGE_HEADER_CONNECTION_NOT_SYNCHRONIZED
from ..data.mrt.bgp.const import BGP_ERROR_MESSAGE_HEADER_BAD_MESSAGE_LENGTH
from ..data.mrt.bgp.const import BGP_ERROR_MESSAGE_HEADER_BAD_MESSAGE_TYPE
from ..data.mrt.bgp.const import BGP_ERROR_OPEN_MESSAGE_UNSPECIFIC
from ..data.mrt.bgp.const import BGP_ERROR_OPEN_MESSAGE_UNSUPPORTED_VERSION_NUMBER
from ..data.mrt.bgp.const import BGP_ERROR_OPEN_MESSAGE_BAD_PEER_AS
from ..data.mrt.bgp.const import BGP_ERROR_OPEN_MESSAGE_BAD_BGP_IDENTIFIER
from ..data.mrt.bgp.const import BGP_ERROR_OPEN_MESSAGE_UNSUPPORTED_OPTIONAL_PARAMETER
from ..data.mrt.bgp.const import BGP_ERROR_OPEN_MESSAGE_AUTHENTICATION_FAILURE
from ..data.mrt.bgp.const import BGP_ERROR_OPEN_MESSAGE_UNACCEPTABLE_HOLD_TIME
from ..data.mrt.bgp.const import BGP_ERROR_OPEN_MESSAGE_UNSUPPORTED_CAPABILITY
from ..data.mrt.bgp.const import BGP_ERROR_OPEN_MESSAGE_DEPRECATED_8
from ..data.mrt.bgp.const import BGP_ERROR_OPEN_MESSAGE_DEPRECATED_9
from ..data.mrt.bgp.const import BGP_ERROR_OPEN_MESSAGE_DEPRECATED_10
from ..data.mrt.bgp.const import BGP_ERROR_OPEN_MESSAGE_ROLE_MISMATCH
from ..data.mrt.bgp.const import BGP_ERROR_UPDATE_MESSAGE_UNSPECIFIC
from ..data.mrt.bgp.const import BGP_ERROR_UPDATE_MESSAGE_MALFORMED_ATTRIBUTE_LIST
from ..data.mrt.bgp.const import BGP_ERROR_UPDATE_MESSAGE_UNRECOGNIZED_WELLKNOWN_ATTRIBUTE
from ..data.mrt.bgp.const import BGP_ERROR_UPDATE_MESSAGE_MISSING_WELLKNOWN_ATTRIBUTE
from ..data.mrt.bgp.const import BGP_ERROR_UPDATE_MESSAGE_INVALID_ATTRIBUTE_FLAGS
from ..data.mrt.bgp.const import BGP_ERROR_UPDATE_MESSAGE_INVALID_ATTRIBUTE_LENGTH
from ..data.mrt.bgp.const import BGP_ERROR_UPDATE_MESSAGE_INVALID_ORIGIN_ATTRIBUTE
from ..data.mrt.bgp.const import BGP_ERROR_UPDATE_MESSAGE_AS_ROUTING_LOOP
from ..data.mrt.bgp.const import BGP_ERROR_UPDATE_MESSAGE_INVALID_NEXTHOP_ATTRIBUTE
from ..data.mrt.bgp.const import BGP_ERROR_UPDATE_MESSAGE_INVALID_OPTIONAL_ATTRIBUTE
from ..data.mrt.bgp.const import BGP_ERROR_UPDATE_MESSAGE_INVALID_NETWORK_FIELD
from ..data.mrt.bgp.const import BGP_ERROR_UPDATE_MESSAGE_MALFORMED_AS_PATH
from ..data.mrt.bgp.const import BGP_ERROR_FSM_UNSPECIFIED
from ..data.mrt.bgp.const import BGP_ERROR_FSM_UNEXPECTED_MESSAGE_IN_OPEN_SENT_STATE
from ..data.mrt.bgp.const import BGP_ERROR_FSM_UNEXPECTED_MESSAGE_IN_OPEN_CONFIRM_STATE
from ..data.mrt.bgp.const import BGP_ERROR_FSM_UNEXPECTED_MESSAGE_IN_ESTABLISHED_STATE
from ..data.mrt.bgp.const import BGP_ERROR_CEASE_RESERVED
from ..data.mrt.bgp.const import BGP_ERROR_CEASE_MAX_NUMBER_PREFIXES_REACHED
from ..data.mrt.bgp.const import BGP_ERROR_CEASE_ADMINISTRATIVE_SHUTDOWN
from ..data.mrt.bgp.const import BGP_ERROR_CEASE_PEER_DECONFIGURED
from ..data.mrt.bgp.const import BGP_ERROR_CEASE_ADMINISTRATIVE_RESET
from ..data.mrt.bgp.const import BGP_ERROR_CEASE_CONNECTION_REJECTED
from ..data.mrt.bgp.const import BGP_ERROR_CEASE_OTHER_CONFIGURATION_CHANGE
from ..data.mrt.bgp.const import BGP_ERROR_CEASE_CONNECTION_COLLISION_RESOLUTION
from ..data.mrt.bgp.const import BGP_ERROR_CEASE_OUT_OF_RESOURCES
from ..data.mrt.bgp.const import BGP_ERROR_CEASE_HARD_RESET
from ..data.mrt.bgp.const import BGP_ERROR_CEASE_BFD_DOWN
from ..data.mrt.bgp.const import BGP_ERROR_ROUTE_REFRESH_RESERVED
from ..data.mrt.bgp.const import BGP_ERROR_ROUTE_REFRESH_INVALID_MESSAGE_LENGTH
from ..data.mrt.bgp.const import BGP_PATH_ATTR_RESERVED
from ..data.mrt.bgp.const import BGP_PATH_ATTR_ORIGIN
from ..data.mrt.bgp.const import BGP_PATH_ATTR_AS_PATH
from ..data.mrt.bgp.const import BGP_PATH_ATTR_NEXT_HOP
from ..data.mrt.bgp.const import BGP_PATH_ATTR_MULTI_EXIT_DISC
from ..data.mrt.bgp.const import BGP_PATH_ATTR_LOCAL_PREF
from ..data.mrt.bgp.const import BGP_PATH_ATTR_ATOMIC_AGGREGATE
from ..data.mrt.bgp.const import BGP_PATH_ATTR_AGGREGATOR
from ..data.mrt.bgp.const import BGP_PATH_ATTR_COMMUNITIES
from ..data.mrt.bgp.const import BGP_PATH_ATTR_ORIGINATOR_ID
from ..data.mrt.bgp.const import BGP_PATH_ATTR_CLUSTER_LIST
from ..data.mrt.bgp.const import BGP_PATH_ATTR_DPA
from ..data.mrt.bgp.const import BGP_PATH_ATTR_ADVERTISER
from ..data.mrt.bgp.const import BGP_PATH_ATTR_RCID_CLUSTER_ID
from ..data.mrt.bgp.const import BGP_PATH_ATTR_MP_REACH_NLRI
from ..data.mrt.bgp.const import BGP_PATH_ATTR_MP_UNREACH_NLRI
from ..data.mrt.bgp.const import BGP_PATH_ATTR_EXTENDED_COMMUNITIES
from ..data.mrt.bgp.const import BGP_PATH_ATTR_AS4_PATH
from ..data.mrt.bgp.const import BGP_PATH_ATTR_AS4_AGGREGATOR
from ..data.mrt.bgp.const import BGP_PATH_ATTR_SAFI_SPECIFIC
from ..data.mrt.bgp.const import BGP_PATH_ATTR_CONNECTOR
from ..data.mrt.bgp.const import BGP_PATH_ATTR_AS_PATHLIMIT
from ..data.mrt.bgp.const import BGP_PATH_ATTR_PMSI_TUNNEL
from ..data.mrt.bgp.const import BGP_PATH_ATTR_TUNNEL_ENCAPSULATION
from ..data.mrt.bgp.const import BGP_PATH_ATTR_TRAFFIC_ENGINEERING
from ..data.mrt.bgp.const import BGP_PATH_ATTR_IPV6_EXTENDED_COMMUNITIES
from ..data.mrt.bgp.const import BGP_PATH_ATTR_AIGP
from ..data.mrt.bgp.const import BGP_PATH_ATTR_PE_DISTINGUISHER_LABELS
from ..data.mrt.bgp.const import BGP_PATH_ATTR_ENTROPY_LABEL_CAPABILITY
from ..data.mrt.bgp.const import BGP_PATH_ATTR_LS
from ..data.mrt.bgp.const import BGP_PATH_ATTR_VENDOR_30
from ..data.mrt.bgp.const import BGP_PATH_ATTR_VENDOR_31
from ..data.mrt.bgp.const import BGP_PATH_ATTR_LARGE_COMMUNITIES
from ..data.mrt.bgp.const import BGP_PATH_ATTR_BGPSEC_PATH
from ..data.mrt.bgp.const import BGP_PATH_ATTR_COMMUNITY_CONTAINER
from ..data.mrt.bgp.const import BGP_PATH_ATTR_ONLY_TO_CUSTOMER
from ..data.mrt.bgp.const import BGP_PATH_ATTR_DOMAIN_PATH
from ..data.mrt.bgp.const import BGP_PATH_ATTR_SFP
from ..data.mrt.bgp.const import BGP_PATH_ATTR_BFD_DISCRIMINATOR
from ..data.mrt.bgp.const import BGP_PATH_ATTR_NHC_TMP
from ..data.mrt.bgp.const import BGP_PATH_ATTR_PREFIX_SID
from ..data.mrt.bgp.const import BGP_PATH_ATTR_ATTR_SET
from ..data.mrt.bgp.const import BGP_PATH_ATTR_VENDOR_129
from ..data.mrt.bgp.const import BGP_PATH_ATTR_VENDOR_241
from ..data.mrt.bgp.const import BGP_PATH_ATTR_VENDOR_242
from ..data.mrt.bgp.const import BGP_PATH_ATTR_VENDOR_243
from ..data.mrt.bgp.const import BGP_PATH_ATTR_RESERVED_FOR_DEV
from ..data.mrt.bgp.const import BGP_PATH_ATTR_ORIGIN_IGP
from ..data.mrt.bgp.const import BGP_PATH_ATTR_ORIGIN_EGP
from ..data.mrt.bgp.const import BGP_PATH_ATTR_ORIGIN_INCOMPLETE

# Main data sources
FTL_MRT = 'mrt'
FTL_LGL = 'lgl'
FTL_BGP = 'bgp'
FTL_PARSER = 'parser'

# Generic record name
FTL_RECORD = 'FTL_RECORD'

# Generic attribute name
FTL_ATTR = 'FTL_ATTR'


def init_const_mappings(**consts):
    """ Generate type-to-string/string-to-type constant mappings.
    """
    # Iterate constants and populate constant mappings
    const_to_str, const_from_str = dict(), dict()
    for const_name, const_value in consts.items():
        const_to_str[const_value] = const_name
        const_from_str[const_name] = const_value

    # Return constant mappings
    return const_to_str, const_from_str


##############################
# BGP STATE_CHANGE CONSTANTS #
##############################

# BGP state_change state constants
(FTL_ATTR_BGP_STATE_CHANGE_STATE_TO_STR,
 FTL_ATTR_BGP_STATE_CHANGE_STATE_FROM_STR) = init_const_mappings(
    idle         = BGP_STATE_IDLE,
    connect      = BGP_STATE_CONNECT,
    active       = BGP_STATE_ACTIVE,
    open_sent    = BGP_STATE_OPEN_SENT,
    open_confirm = BGP_STATE_OPEN_CONFIRM,
    established  = BGP_STATE_ESTABLISHED,
    clearing     = BGP_STATE_CLEARING,
    deleted      = BGP_STATE_DELETED,
)

#######################
# BGP ROUTE CONSTANTS #
#######################

# BGP route source integer types
FTL_ATTR_BGP_ROUTE_SOURCE_RIB      = 0
FTL_ATTR_BGP_ROUTE_SOURCE_ANNOUNCE = 1
FTL_ATTR_BGP_ROUTE_SOURCE_WITHDRAW = 2

# BGP route source string types
FTL_ATTR_BGP_ROUTE_SOURCE_RIB_STR      = 'rib'
FTL_ATTR_BGP_ROUTE_SOURCE_ANNOUNCE_STR = 'announce'
FTL_ATTR_BGP_ROUTE_SOURCE_WITHDRAW_STR = 'withdraw'

# BGP route source constants
(FTL_ATTR_BGP_ROUTE_SOURCE_TO_STR,
 FTL_ATTR_BGP_ROUTE_SOURCE_FROM_STR) = init_const_mappings(**{
    FTL_ATTR_BGP_ROUTE_SOURCE_RIB_STR:      FTL_ATTR_BGP_ROUTE_SOURCE_RIB,
    FTL_ATTR_BGP_ROUTE_SOURCE_ANNOUNCE_STR: FTL_ATTR_BGP_ROUTE_SOURCE_ANNOUNCE,
    FTL_ATTR_BGP_ROUTE_SOURCE_WITHDRAW_STR: FTL_ATTR_BGP_ROUTE_SOURCE_WITHDRAW,
})

# BGP route origin integer types
FTL_ATTR_BGP_ROUTE_ORIGIN_IGP        = BGP_PATH_ATTR_ORIGIN_IGP
FTL_ATTR_BGP_ROUTE_ORIGIN_EGP        = BGP_PATH_ATTR_ORIGIN_EGP
FTL_ATTR_BGP_ROUTE_ORIGIN_INCOMPLETE = BGP_PATH_ATTR_ORIGIN_INCOMPLETE

# BGP route origin string types
FTL_ATTR_BGP_ROUTE_ORIGIN_IGP_STR        = 'igp'
FTL_ATTR_BGP_ROUTE_ORIGIN_EGP_STR        = 'egp'
FTL_ATTR_BGP_ROUTE_ORIGIN_INCOMPLETE_STR = 'incomplete'

# BGP route origin constants
(FTL_ATTR_BGP_ROUTE_ORIGIN_TO_STR,
 FTL_ATTR_BGP_ROUTE_ORIGIN_FROM_STR) = init_const_mappings(**{
    FTL_ATTR_BGP_ROUTE_ORIGIN_IGP_STR:        BGP_PATH_ATTR_ORIGIN_IGP,
    FTL_ATTR_BGP_ROUTE_ORIGIN_EGP_STR:        BGP_PATH_ATTR_ORIGIN_EGP,
    FTL_ATTR_BGP_ROUTE_ORIGIN_INCOMPLETE_STR: BGP_PATH_ATTR_ORIGIN_INCOMPLETE,
})

##############################
# BGP NOTIFICATION CONSTANTS #
##############################

# BGP notification error code constants
(FTL_ATTR_BGP_NOTIFICATION_ERROR_CODE_TO_STR,
 FTL_ATTR_BGP_NOTIFICATION_ERROR_CODE_FROM_STR) = init_const_mappings(
    reserved           = BGP_ERROR_RESERVED,
    message_header     = BGP_ERROR_MESSAGE_HEADER,
    open_message       = BGP_ERROR_OPEN_MESSAGE,
    update_message     = BGP_ERROR_UPDATE_MESSAGE,
    hold_timer_expired = BGP_ERROR_HOLD_TIMER_EXPIRED,
    fsm                = BGP_ERROR_FSM,
    cease              = BGP_ERROR_CEASE,
    route_refresh      = BGP_ERROR_ROUTE_REFRESH,
)

# BGP message header error subcode constants
(FTL_ATTR_BGP_NOTIFICATION_ERROR_MESSAGE_HEADER_TO_STR,
 FTL_ATTR_BGP_NOTIFICATION_ERROR_MESSAGE_HEADER_FROM_STR) = init_const_mappings(
    unspecific                  = BGP_ERROR_MESSAGE_HEADER_UNSPECIFIC,
    connection_not_synchronized = BGP_ERROR_MESSAGE_HEADER_CONNECTION_NOT_SYNCHRONIZED,
    message_length              = BGP_ERROR_MESSAGE_HEADER_BAD_MESSAGE_LENGTH,
    message_tgype               = BGP_ERROR_MESSAGE_HEADER_BAD_MESSAGE_TYPE,
)

# BGP OPEN message error subcode constants
(FTL_ATTR_BGP_NOTIFICATION_ERROR_OPEN_MESSAGE_TO_STR,
 FTL_ATTR_BGP_NOTIFICATION_ERROR_OPEN_MESSAGE_FROM_STR) = init_const_mappings(
    unspecific                     = BGP_ERROR_OPEN_MESSAGE_UNSPECIFIC,
    unsupported_version_number     = BGP_ERROR_OPEN_MESSAGE_UNSUPPORTED_VERSION_NUMBER,
    bad_peer_as                    = BGP_ERROR_OPEN_MESSAGE_BAD_PEER_AS,
    bad_bgp_identifier             = BGP_ERROR_OPEN_MESSAGE_BAD_BGP_IDENTIFIER,
    unsupported_optional_parameter = BGP_ERROR_OPEN_MESSAGE_UNSUPPORTED_OPTIONAL_PARAMETER,
    authentication_failure         = BGP_ERROR_OPEN_MESSAGE_AUTHENTICATION_FAILURE,
    unacceptable_hold_time         = BGP_ERROR_OPEN_MESSAGE_UNACCEPTABLE_HOLD_TIME,
    unsupported_capability         = BGP_ERROR_OPEN_MESSAGE_UNSUPPORTED_CAPABILITY,
    deprecated_8                   = BGP_ERROR_OPEN_MESSAGE_DEPRECATED_8,
    deprecated_9                   = BGP_ERROR_OPEN_MESSAGE_DEPRECATED_9,
    deprecated_10                  = BGP_ERROR_OPEN_MESSAGE_DEPRECATED_10,
    role_mismatch                  = BGP_ERROR_OPEN_MESSAGE_ROLE_MISMATCH,
)

# BGP UPDATE message error subcode constants
(FTL_ATTR_BGP_NOTIFICATION_ERROR_UPDATE_MESSAGE_TO_STR,
 FTL_ATTR_BGP_NOTIFICATION_ERROR_UPDATE_MESSAGE_FROM_STR) = init_const_mappings(
    unspecific                       = BGP_ERROR_UPDATE_MESSAGE_UNSPECIFIC,
    malformed_attribute_list         = BGP_ERROR_UPDATE_MESSAGE_MALFORMED_ATTRIBUTE_LIST,
    unrecognized_wellknown_attribute = BGP_ERROR_UPDATE_MESSAGE_UNRECOGNIZED_WELLKNOWN_ATTRIBUTE,
    missing_wellknown_attribute      = BGP_ERROR_UPDATE_MESSAGE_MISSING_WELLKNOWN_ATTRIBUTE,
    invalid_attribute_flags          = BGP_ERROR_UPDATE_MESSAGE_INVALID_ATTRIBUTE_FLAGS,
    invalid_attribute_length         = BGP_ERROR_UPDATE_MESSAGE_INVALID_ATTRIBUTE_LENGTH,
    invalid_origin_attribute         = BGP_ERROR_UPDATE_MESSAGE_INVALID_ORIGIN_ATTRIBUTE,
    as_routing_loop                  = BGP_ERROR_UPDATE_MESSAGE_AS_ROUTING_LOOP,
    invalid_nexthop_attribute        = BGP_ERROR_UPDATE_MESSAGE_INVALID_NEXTHOP_ATTRIBUTE,
    invalid_optional_attribute       = BGP_ERROR_UPDATE_MESSAGE_INVALID_OPTIONAL_ATTRIBUTE,
    invalid_network_field            = BGP_ERROR_UPDATE_MESSAGE_INVALID_NETWORK_FIELD,
    malformed_as_path                = BGP_ERROR_UPDATE_MESSAGE_MALFORMED_AS_PATH,
)

# BGP finite state machine error subcode constants
(FTL_ATTR_BGP_NOTIFICATION_ERROR_FSM_TO_STR,
 FTL_ATTR_BGP_NOTIFICATION_ERROR_FSM_FROM_STR) = init_const_mappings(
    unspecified                              = BGP_ERROR_FSM_UNSPECIFIED,
    unexpected_message_in_open_sent_state    = BGP_ERROR_FSM_UNEXPECTED_MESSAGE_IN_OPEN_SENT_STATE,
    unexpected_message_in_open_confirm_state = BGP_ERROR_FSM_UNEXPECTED_MESSAGE_IN_OPEN_CONFIRM_STATE,
    unexpected_message_in_established_state  = BGP_ERROR_FSM_UNEXPECTED_MESSAGE_IN_ESTABLISHED_STATE,
)

# BGP cease error subcode constants
(FTL_ATTR_BGP_NOTIFICATION_ERROR_CEASE_TO_STR,
 FTL_ATTR_BGP_NOTIFICATION_ERROR_CEASE_FROM_STR) = init_const_mappings(
    reserved                        = BGP_ERROR_CEASE_RESERVED,
    max_number_prefixes_reached     = BGP_ERROR_CEASE_MAX_NUMBER_PREFIXES_REACHED,
    administrative_shutdown         = BGP_ERROR_CEASE_ADMINISTRATIVE_SHUTDOWN,
    peer_deconfigured               = BGP_ERROR_CEASE_PEER_DECONFIGURED,
    administrative_reset            = BGP_ERROR_CEASE_ADMINISTRATIVE_RESET,
    connection_rejected             = BGP_ERROR_CEASE_CONNECTION_REJECTED,
    other_configuration_change      = BGP_ERROR_CEASE_OTHER_CONFIGURATION_CHANGE,
    connection_collision_resolution = BGP_ERROR_CEASE_CONNECTION_COLLISION_RESOLUTION,
    out_of_resources                = BGP_ERROR_CEASE_OUT_OF_RESOURCES,
    hard_reset                      = BGP_ERROR_CEASE_HARD_RESET,
    bfd_down                        = BGP_ERROR_CEASE_BFD_DOWN,
)

# BGP route refresh error subcode constants
(FTL_ATTR_BGP_NOTIFICATION_ERROR_ROUTE_REFRESH_TO_STR,
 FTL_ATTR_BGP_NOTIFICATION_ERROR_ROUTE_REFRESH_FROM_STR) = init_const_mappings(
    refresh_reserved               = BGP_ERROR_ROUTE_REFRESH_RESERVED,
    refresh_invalid_message_length = BGP_ERROR_ROUTE_REFRESH_INVALID_MESSAGE_LENGTH,
)

# BGP notification code/subcode code-to-constant map
FTL_ATTR_BGP_NOTIFICATION_ERROR_SUBCODE_TO_STR = {
    BGP_ERROR_MESSAGE_HEADER: FTL_ATTR_BGP_NOTIFICATION_ERROR_MESSAGE_HEADER_TO_STR,
    BGP_ERROR_OPEN_MESSAGE:   FTL_ATTR_BGP_NOTIFICATION_ERROR_OPEN_MESSAGE_TO_STR,
    BGP_ERROR_UPDATE_MESSAGE: FTL_ATTR_BGP_NOTIFICATION_ERROR_UPDATE_MESSAGE_TO_STR,
    BGP_ERROR_FSM:            FTL_ATTR_BGP_NOTIFICATION_ERROR_FSM_TO_STR,
    BGP_ERROR_CEASE:          FTL_ATTR_BGP_NOTIFICATION_ERROR_CEASE_TO_STR,
    BGP_ERROR_ROUTE_REFRESH:  FTL_ATTR_BGP_NOTIFICATION_ERROR_ROUTE_REFRESH_TO_STR,
}

# BGP notification code/subcode constant-to-code map
FTL_ATTR_BGP_NOTIFICATION_ERROR_SUBCODE_FROM_STR = dict(
    message_header = FTL_ATTR_BGP_NOTIFICATION_ERROR_MESSAGE_HEADER_FROM_STR,
    open_message   = FTL_ATTR_BGP_NOTIFICATION_ERROR_OPEN_MESSAGE_FROM_STR,
    update_message = FTL_ATTR_BGP_NOTIFICATION_ERROR_UPDATE_MESSAGE_FROM_STR,
    fsm            = FTL_ATTR_BGP_NOTIFICATION_ERROR_FSM_FROM_STR,
    cease          = FTL_ATTR_BGP_NOTIFICATION_ERROR_CEASE_FROM_STR,
    route_refresh  = FTL_ATTR_BGP_NOTIFICATION_ERROR_ROUTE_REFRESH_FROM_STR,
)

######################
# BGP OPEN CONSTANTS #
######################

# BGP open capability types
(FTL_ATTR_BGP_OPEN_CAPABILITY_TO_STR,
 FTL_ATTR_BGP_OPEN_CAPABILITY_FROM_STR) = init_const_mappings(
    reserved_0                 = BGP_CAPABILITY_RESERVED_0,
    bgp4mp                     = BGP_CAPABILITY_BGP4MP,
    route_refresh              = BGP_CAPABILITY_ROUTE_REFRESH,
    outbound_filter            = BGP_CAPABILITY_OUTBOUND_FILTER,
    multiple_routes            = BGP_CAPABILITY_MULTIPLE_ROUTES,
    extended_next_hop          = BGP_CAPABILITY_EXTENDED_NEXT_HOP,
    bgp4mp_et                  = BGP_CAPABILITY_BGP4MP_ET,
    bgpsec                     = BGP_CAPABILITY_BGPSEC,
    multiple_labels            = BGP_CAPABILITY_MULTIPLE_LABELS,
    bgp_role                   = BGP_CAPABILITY_BGP_ROLE,
    graceful_restart           = BGP_CAPABILITY_GRACEFUL_RESTART,
    as4                        = BGP_CAPABILITY_AS4,
    unknown_66                 = BGP_CAPABILITY_UNKNOWN_66,
    dynamic                    = BGP_CAPABILITY_DYNAMIC,
    multisession               = BGP_CAPABILITY_MULTISESSION,
    addpath                    = BGP_CAPABILITY_ADDPATH,
    enhanced_route_refresh     = BGP_CAPABILITY_ENHANCED_ROUTE_REFRESH,
    llgr                       = BGP_CAPABILITY_LLGR,
    policy_distribution        = BGP_CAPABILITY_POLICY_DISTRIBUTION,
    fqdn                       = BGP_CAPABILITY_FQDN,
    bfd                        = BGP_CAPABILITY_BFD,
    software_version           = BGP_CAPABILITY_SOFTWARE_VERSION,
    prestd_route_refresh       = BGP_CAPABILITY_PRESTD_ROUTE_REFRESH,
    prestd_policy_distribution = BGP_CAPABILITY_PRESTD_POLICY_DISTRIBUTION,
    prestd_outbound_filter     = BGP_CAPABILITY_PRESTD_OUTBOUND_FILTER,
    prestd_multisession        = BGP_CAPABILITY_PRESTD_MULTISESSION,
    prestd_fqdn                = BGP_CAPABILITY_PRESTD_FQDN,
    prestd_operational         = BGP_CAPABILITY_PRESTD_OPERATIONAL,
    reserved_255               = BGP_CAPABILITY_RESERVED_255,
)

#######################
# BGP STATS CONSTANTS #
#######################

# BGP stats MRT fixes integer types
FTL_ATTR_BGP_STATS_MRT_FIXES_BGP4MP_ADDPATH  = 0
FTL_ATTR_BGP_STATS_MRT_FIXES_BGP4MP_ASLENGTH = 1
FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_ADDPATH    = 2
FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_ASLENGTH   = 3
FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_NLRI       = 4

# BGP stats MRT fixes string types
FTL_ATTR_BGP_STATS_MRT_FIXES_BGP4MP_ADDPATH_STR  = 'bgp4mp_addpath'
FTL_ATTR_BGP_STATS_MRT_FIXES_BGP4MP_ASLENGTH_STR = 'bgp4mp_aslength'
FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_ADDPATH_STR    = 'tdv2_addpath'
FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_ASLENGTH_STR   = 'tdv2_aslength'
FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_NLRI_STR       = 'tdv2_nlri'

# BGP stats MRT fixes constants
(FTL_ATTR_BGP_STATS_MRT_FIXES_TO_STR,
 FTL_ATTR_BGP_STATS_MRT_FIXES_FROM_STR) = init_const_mappings(**{
    FTL_ATTR_BGP_STATS_MRT_FIXES_BGP4MP_ADDPATH_STR: FTL_ATTR_BGP_STATS_MRT_FIXES_BGP4MP_ADDPATH,
    FTL_ATTR_BGP_STATS_MRT_FIXES_BGP4MP_ASLENGTH_STR: FTL_ATTR_BGP_STATS_MRT_FIXES_BGP4MP_ASLENGTH,
    FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_ADDPATH_STR: FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_ADDPATH,
    FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_ASLENGTH_STR: FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_ASLENGTH,
    FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_NLRI_STR: FTL_ATTR_BGP_STATS_MRT_FIXES_TDV2_NLRI,
})

# BGP stats MRT entry type constants
(FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_TYPE_TO_STR,
 FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_TYPE_FROM_STR) = init_const_mappings(
    null          = MRT_NULL,
    start         = MRT_START,
    die           = MRT_DIE,
    i_am_dead     = MRT_I_AM_DEAD,
    peer_down     = MRT_PEER_DOWN,
    bgp           = MRT_BGP,
    rip           = MRT_RIP,
    idrp          = MRT_IDRP,
    ripng         = MRT_RIPNG,
    bgp4plus      = MRT_BGP4PLUS,
    bgp4plus_01   = MRT_BGP4PLUS_01,
    ospfv2        = MRT_OSPFV2,
    table_dump    = MRT_TABLE_DUMP,
    table_dump_v2 = MRT_TABLE_DUMP_V2,
    bgp4mp        = MRT_BGP4MP,
    bgp4mp_et     = MRT_BGP4MP_ET,
    isis          = MRT_ISIS,
    isis_et       = MRT_ISIS_ET,
    ospfv3        = MRT_OSPFV3,
    ospfv3_et     = MRT_OSPFV3_ET,
)

# BGP stats MRT entry subtype BGP4MP constants
(FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_BGP4MP_TO_STR,
 FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_BGP4MP_FROM_STR) = init_const_mappings(
    entry                     = MRT_BGP4MP_ENTRY,
    snapshot                  = MRT_BGP4MP_SNAPSHOT,
    state_change              = MRT_BGP4MP_STATE_CHANGE,
    state_change_as4          = MRT_BGP4MP_STATE_CHANGE_AS4,
    message                   = MRT_BGP4MP_MESSAGE,
    message_addpath           = MRT_BGP4MP_MESSAGE_ADDPATH,
    message_local             = MRT_BGP4MP_MESSAGE_LOCAL,
    message_local_addpath     = MRT_BGP4MP_MESSAGE_LOCAL_ADDPATH,
    message_as4               = MRT_BGP4MP_MESSAGE_AS4,
    message_as4_addpath       = MRT_BGP4MP_MESSAGE_AS4_ADDPATH,
    message_as4_local         = MRT_BGP4MP_MESSAGE_AS4_LOCAL,
    message_as4_local_addpath = MRT_BGP4MP_MESSAGE_AS4_LOCAL_ADDPATH,
)

# BGP stats MRT entry subtype BGP4MP_ET constants
FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_BGP4MP_ET_TO_STR = FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_BGP4MP_TO_STR
FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_BGP4MP_ET_FROM_STR = FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_BGP4MP_FROM_STR

# BGP stats MRT entry subtype table dump v2 constants
(FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_TABLE_DUMP_V2_TO_STR,
 FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_TABLE_DUMP_V2_FROM_STR) = init_const_mappings(
    peer_index_table           = MRT_TABLE_DUMP_V2_PEER_INDEX_TABLE,
    geo_peer_table             = MRT_TABLE_DUMP_V2_GEO_PEER_TABLE,
    rib_generic                = MRT_TABLE_DUMP_V2_RIB_GENERIC,
    rib_generic_addpath        = MRT_TABLE_DUMP_V2_RIB_GENERIC_ADDPATH,
    rib_ipv4_unicast           = MRT_TABLE_DUMP_V2_RIB_IPV4_UNICAST,
    rib_ipv4_multicast         = MRT_TABLE_DUMP_V2_RIB_IPV4_MULTICAST,
    rib_ipv6_unicast           = MRT_TABLE_DUMP_V2_RIB_IPV6_UNICAST,
    rib_ipv6_multicast         = MRT_TABLE_DUMP_V2_RIB_IPV6_MULTICAST,
    rib_ipv4_unicast_addpath   = MRT_TABLE_DUMP_V2_RIB_IPV4_UNICAST_ADDPATH,
    rib_ipv4_multicast_addpath = MRT_TABLE_DUMP_V2_RIB_IPV4_MULTICAST_ADDPATH,
    rib_ipv6_unicast_addpath   = MRT_TABLE_DUMP_V2_RIB_IPV6_UNICAST_ADDPATH,
    rib_ipv6_multicast_addpath = MRT_TABLE_DUMP_V2_RIB_IPV6_MULTICAST_ADDPATH,
)

# BGP stats MRT entry subtype BGP constants
(FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_BGP_TO_STR,
 FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_BGP_FROM_STR) = init_const_mappings(
    null         = MRT_BGP_NULL,
    update       = MRT_BGP_UPDATE,
    pref_update  = MRT_BGP_PREF_UPDATE,
    state_change = MRT_BGP_STATE_CHANGE,
    sync         = MRT_BGP_SYNC,
    open         = MRT_BGP_OPEN,
    notify       = MRT_BGP_NOTIFY,
    keepalive    = MRT_BGP_KEEPALIVE,
)

 # BGP stats MRT entry subtype table dump constants
(FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_TABLE_DUMP_TO_STR,
 FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_TABLE_DUMP_FROM_STR) = init_const_mappings(
    afi_ipv4 = AFI_IPV4,
    afi_ipv6 = AFI_IPV6,
)

# BGP stats MRT entry type/subtype type-to-constant map
FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_TO_STR = {
    MRT_BGP4MP:        FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_BGP4MP_TO_STR,
    MRT_BGP4MP_ET:     FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_BGP4MP_ET_TO_STR,
    MRT_TABLE_DUMP_V2: FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_TABLE_DUMP_V2_TO_STR,
    MRT_BGP:           FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_BGP_TO_STR,
    MRT_TABLE_DUMP:    FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_TABLE_DUMP_TO_STR,
}

# BGP stats MRT entry type/subtype constant-to-type map
FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_FROM_STR = dict(
    bgp4mp        = FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_BGP4MP_FROM_STR,
    bgp4mp_et     = FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_BGP4MP_ET_FROM_STR,
    table_dump_v2 = FTL_ATTR_BGP_STATS_MRT_BGP_ENTRY_SUBTYPE_TABLE_DUMP_V2_FROM_STR,
)

# BGP stats BGP message type constants
(FTL_ATTR_BGP_STATS_BGP_MESSAGE_TYPE_TO_STR,
 FTL_ATTR_BGP_STATS_BGP_MESSAGE_TYPE_FROM_STR) = init_const_mappings(
    reserved      = BGP_BGP4MP_RESERVED,
    open          = BGP_BGP4MP_OPEN,
    update        = BGP_BGP4MP_UPDATE,
    notification  = BGP_BGP4MP_NOTIFICATION,
    keepalive     = BGP_BGP4MP_KEEPALIVE,
    route_refresh = BGP_BGP4MP_ROUTE_REFRESH,
)

# BGP stats capability type constants
FTL_ATTR_BGP_STATS_BGP_CAPABILITY_TYPE_TO_STR = FTL_ATTR_BGP_OPEN_CAPABILITY_TO_STR
FTL_ATTR_BGP_STATS_BGP_CAPABILITY_TYPE_FROM_STR = FTL_ATTR_BGP_OPEN_CAPABILITY_FROM_STR

# BGP stats BGP path attribute constants
(FTL_ATTR_BGP_STATS_BGP_PATH_ATTR_TO_STR,
 FTL_ATTR_BGP_STATS_BGP_PATH_ATTR_FORM_STR) = init_const_mappings(
    reserved                  = BGP_PATH_ATTR_RESERVED,
    origin                    = BGP_PATH_ATTR_ORIGIN,
    as_path                   = BGP_PATH_ATTR_AS_PATH,
    next_hop                  = BGP_PATH_ATTR_NEXT_HOP,
    multi_exit_disc           = BGP_PATH_ATTR_MULTI_EXIT_DISC,
    local_pref                = BGP_PATH_ATTR_LOCAL_PREF,
    atomic_aggregate          = BGP_PATH_ATTR_ATOMIC_AGGREGATE,
    aggregator                = BGP_PATH_ATTR_AGGREGATOR,
    communities               = BGP_PATH_ATTR_COMMUNITIES,
    originator_id             = BGP_PATH_ATTR_ORIGINATOR_ID,
    cluster_list              = BGP_PATH_ATTR_CLUSTER_LIST,
    dpa                       = BGP_PATH_ATTR_DPA,
    advertiser                = BGP_PATH_ATTR_ADVERTISER,
    rcid_cluster_id           = BGP_PATH_ATTR_RCID_CLUSTER_ID,
    mp_reach_nlri             = BGP_PATH_ATTR_MP_REACH_NLRI,
    mp_unreach_nlri           = BGP_PATH_ATTR_MP_UNREACH_NLRI,
    extended_communities      = BGP_PATH_ATTR_EXTENDED_COMMUNITIES,
    as4_path                  = BGP_PATH_ATTR_AS4_PATH,
    as4_aggregator            = BGP_PATH_ATTR_AS4_AGGREGATOR,
    safi_specific             = BGP_PATH_ATTR_SAFI_SPECIFIC,
    connector                 = BGP_PATH_ATTR_CONNECTOR,
    as_pathlimit              = BGP_PATH_ATTR_AS_PATHLIMIT,
    pmsi_tunnel               = BGP_PATH_ATTR_PMSI_TUNNEL,
    tunnel_encapsulation      = BGP_PATH_ATTR_TUNNEL_ENCAPSULATION,
    traffic_engineering       = BGP_PATH_ATTR_TRAFFIC_ENGINEERING,
    ipv6_extended_communities = BGP_PATH_ATTR_IPV6_EXTENDED_COMMUNITIES,
    aigp                      = BGP_PATH_ATTR_AIGP,
    pe_distinguisher_labels   = BGP_PATH_ATTR_PE_DISTINGUISHER_LABELS,
    entropy_label_capability  = BGP_PATH_ATTR_ENTROPY_LABEL_CAPABILITY,
    ls                        = BGP_PATH_ATTR_LS,
    vendor_30                 = BGP_PATH_ATTR_VENDOR_30,
    vendor_31                 = BGP_PATH_ATTR_VENDOR_31,
    large_communities         = BGP_PATH_ATTR_LARGE_COMMUNITIES,
    bgpsec_path               = BGP_PATH_ATTR_BGPSEC_PATH,
    community_container       = BGP_PATH_ATTR_COMMUNITY_CONTAINER,
    only_to_customer          = BGP_PATH_ATTR_ONLY_TO_CUSTOMER,
    domain_path               = BGP_PATH_ATTR_DOMAIN_PATH,
    sfp                       = BGP_PATH_ATTR_SFP,
    bfd_discriminator         = BGP_PATH_ATTR_BFD_DISCRIMINATOR,
    nhc_tmp                   = BGP_PATH_ATTR_NHC_TMP,
    prefix_sid                = BGP_PATH_ATTR_PREFIX_SID,
    attr_set                  = BGP_PATH_ATTR_ATTR_SET,
    vendor_129                = BGP_PATH_ATTR_VENDOR_129,
    vendor_241                = BGP_PATH_ATTR_VENDOR_241,
    vendor_242                = BGP_PATH_ATTR_VENDOR_242,
    vendor_243                = BGP_PATH_ATTR_VENDOR_243,
    reserved_for_dev          = BGP_PATH_ATTR_RESERVED_FOR_DEV,
)

#######################
# BGP ERROR CONSTANTS #
#######################

# BGP error source integer types
FTL_ATTR_BGP_ERROR_SOURCE_MRT    = 0
FTL_ATTR_BGP_ERROR_SOURCE_LGL    = 1
FTL_ATTR_BGP_ERROR_SOURCE_PARSER = 2

# BGP error source string types
FTL_ATTR_BGP_ERROR_SOURCE_MRT_STR    = FTL_MRT
FTL_ATTR_BGP_ERROR_SOURCE_LGL_STR    = FTL_LGL
FTL_ATTR_BGP_ERROR_SOURCE_PARSER_STR = FTL_PARSER

# BGP error source constants
(FTL_ATTR_BGP_ERROR_SOURCE_TO_STR,
 FTL_ATTR_BGP_ERROR_SOURCE_FROM_STR) = init_const_mappings(**{
    FTL_ATTR_BGP_ERROR_SOURCE_MRT_STR:    FTL_ATTR_BGP_ERROR_SOURCE_MRT,
    FTL_ATTR_BGP_ERROR_SOURCE_LGL_STR:    FTL_ATTR_BGP_ERROR_SOURCE_LGL,
    FTL_ATTR_BGP_ERROR_SOURCE_PARSER_STR: FTL_ATTR_BGP_ERROR_SOURCE_PARSER,
})

# BGP error scope integer types
FTL_ATTR_BGP_ERROR_SCOPE_BASE   = 0
FTL_ATTR_BGP_ERROR_SCOPE_FILE   = 1
FTL_ATTR_BGP_ERROR_SCOPE_HEADER = 2
FTL_ATTR_BGP_ERROR_SCOPE_FORMAT = 3
FTL_ATTR_BGP_ERROR_SCOPE_DATA   = 4

# BGP error scope string types
FTL_ATTR_BGP_ERROR_SCOPE_BASE_STR   = 'base'
FTL_ATTR_BGP_ERROR_SCOPE_FILE_STR   = 'file'
FTL_ATTR_BGP_ERROR_SCOPE_HEADER_STR = 'header'
FTL_ATTR_BGP_ERROR_SCOPE_FORMAT_STR = 'format'
FTL_ATTR_BGP_ERROR_SCOPE_DATA_STR   = 'data'

# BGP error scope constants
(FTL_ATTR_BGP_ERROR_SCOPE_TO_STR,
 FTL_ATTR_BGP_ERROR_SCOPE_FROM_STR) = init_const_mappings(**{
    FTL_ATTR_BGP_ERROR_SCOPE_BASE_STR:   FTL_ATTR_BGP_ERROR_SCOPE_BASE,
    FTL_ATTR_BGP_ERROR_SCOPE_FILE_STR:   FTL_ATTR_BGP_ERROR_SCOPE_FILE,
    FTL_ATTR_BGP_ERROR_SCOPE_HEADER_STR: FTL_ATTR_BGP_ERROR_SCOPE_HEADER,
    FTL_ATTR_BGP_ERROR_SCOPE_FORMAT_STR: FTL_ATTR_BGP_ERROR_SCOPE_FORMAT,
    FTL_ATTR_BGP_ERROR_SCOPE_DATA_STR:   FTL_ATTR_BGP_ERROR_SCOPE_DATA,
})

# BGP error reason integer types
FTL_ATTR_BGP_ERROR_REASON_RUNTIME          = 0
FTL_ATTR_BGP_ERROR_REASON_MISSING_DATA     = 1
FTL_ATTR_BGP_ERROR_REASON_INVALID_TYPE     = 2
FTL_ATTR_BGP_ERROR_REASON_INVALID_ATTR     = 3
FTL_ATTR_BGP_ERROR_REASON_INVALID_PROTOCOL = 4
FTL_ATTR_BGP_ERROR_REASON_INVALID_PREFIX   = 5
FTL_ATTR_BGP_ERROR_REASON_INVALID_ASPATH   = 6
FTL_ATTR_BGP_ERROR_REASON_INVALID_AS       = 7
FTL_ATTR_BGP_ERROR_REASON_INVALID_IP       = 8

# BGP error reason string types
FTL_ATTR_BGP_ERROR_REASON_RUNTIME_STR          = 'runtime'
FTL_ATTR_BGP_ERROR_REASON_MISSING_DATA_STR     = 'missing_data'
FTL_ATTR_BGP_ERROR_REASON_INVALID_TYPE_STR     = 'invalid_type'
FTL_ATTR_BGP_ERROR_REASON_INVALID_ATTR_STR     = 'invalid_attr'
FTL_ATTR_BGP_ERROR_REASON_INVALID_PROTOCOL_STR = 'invalid_protocol'
FTL_ATTR_BGP_ERROR_REASON_INVALID_PREFIX_STR   = 'invalid_prefix'
FTL_ATTR_BGP_ERROR_REASON_INVALID_ASPATH_STR   = 'invalid_aspath'
FTL_ATTR_BGP_ERROR_REASON_INVALID_AS_STR       = 'invalid_as'
FTL_ATTR_BGP_ERROR_REASON_INVALID_IP_STR       = 'invalid_ip'

# BGP error reason constants
(FTL_ATTR_BGP_ERROR_REASON_TO_STR,
 FTL_ATTR_BGP_ERROR_REASON_FROM_STR) = init_const_mappings(**{
    FTL_ATTR_BGP_ERROR_REASON_RUNTIME_STR:          FTL_ATTR_BGP_ERROR_REASON_RUNTIME,
    FTL_ATTR_BGP_ERROR_REASON_MISSING_DATA_STR:     FTL_ATTR_BGP_ERROR_REASON_MISSING_DATA,
    FTL_ATTR_BGP_ERROR_REASON_INVALID_TYPE_STR:     FTL_ATTR_BGP_ERROR_REASON_INVALID_TYPE,
    FTL_ATTR_BGP_ERROR_REASON_INVALID_ATTR_STR:     FTL_ATTR_BGP_ERROR_REASON_INVALID_ATTR,
    FTL_ATTR_BGP_ERROR_REASON_INVALID_PROTOCOL_STR: FTL_ATTR_BGP_ERROR_REASON_INVALID_PROTOCOL,
    FTL_ATTR_BGP_ERROR_REASON_INVALID_PREFIX_STR:   FTL_ATTR_BGP_ERROR_REASON_INVALID_PREFIX,
    FTL_ATTR_BGP_ERROR_REASON_INVALID_ASPATH_STR:   FTL_ATTR_BGP_ERROR_REASON_INVALID_ASPATH,
    FTL_ATTR_BGP_ERROR_REASON_INVALID_AS_STR:       FTL_ATTR_BGP_ERROR_REASON_INVALID_AS,
    FTL_ATTR_BGP_ERROR_REASON_INVALID_IP_STR:       FTL_ATTR_BGP_ERROR_REASON_INVALID_IP,
})



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
# flake8: noqa
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

###################
# BGP SAFI VALUES #
###################

BGP_SAFI_RESERVED        = 0    # Defined in RFC4760
BGP_SAFI_UNICAST         = 1    # Defined in RFC4760
BGP_SAFI_MULTICAST       = 2    # Defined in RFC4760
BGP_SAFI_BOTH            = 3    # Deprecated in RFC4760
BGP_SAFI_MPLS            = 4    # Defined in RFC8277
BGP_SAFI_MCAST_VPN       = 5    # Defined in RFC6514
BGP_SAFI_PSEUDOWIRE      = 6    # Defined in RFC7267
BGP_SAFI_ENCAPSULATION   = 7    # Deprecated in RFC9012
BGP_SAFI_MCAST_VPLS      = 8    # Defined in RFC7117
BGP_SAFI_BGP_SFC         = 9    # Defined in RFC9015
BGP_SAFI_TUNNEL          = 64   # Proposed in draft_nalawade_kapoor_tunnel_safi
BGP_SAFI_VPLS            = 65   # Defined in RFC4761
BGP_SAFI_MDT             = 66   # Defined in RFC6037
BGP_SAFI_4OVER6          = 67   # Defined in RFC5747
BGP_SAFI_6OVER4          = 68   # Proposed by Yong Cui
BGP_SAFI_L1VPN           = 69   # Defined in RFC5195
BGP_SAFI_EVPN            = 70   # Defined in RFC7432
BGP_SAFI_LS              = 71   # Defined in RFC9552
BGP_SAFI_LS_VPN          = 72   # Defined in RFC9552
BGP_SAFI_SR_TE_POLICY    = 73   # Proposed in draft_ietf_idr_segment_routing_te_policy
BGP_SAFI_SD_WAN          = 74   # Proposed in draft_ietf_idr_sdwan_edge_discovery
BGP_SAFI_NG_POLICY       = 75   # Proposed in draft_ietf_idr_rpd
BGP_SAFI_CT              = 76   # Proposed in draft_ietf_idr_ct
BGP_SAFI_TUNNEL_FLOWSPEC = 77   # Proposed in draft_ietf_idr_flowspec_nvo3
BGP_SAFI_MCAST_TREE      = 78   # Proposed in draft_ietf_bess_multicast
BGP_SAFI_DPS             = 79   # Proposed by https://eos.arista.com/eos_4_26_2f/dps_vpn_scaling_using_bgp
BGP_SAFI_LSVR_SPF        = 80   # Proposed in draft_ietf_lsvr_spf
BGP_SAFI_CAR             = 83   # Proposed in draft_ietf_idr_car
BGP_SAFI_VPN_CAR         = 84   # Proposed in draft_ietf_idr_car
BGP_SAFI_MUP             = 85   # Proposed in draft_mpmz_bess_mup_safi
BGP_SAFI_L3VPN           = 128  # Defined in RFC4364
BGP_SAFI_L3VPN_MULTICAST = 129  # Defined in RFC6513
BGP_SAFI_ROUTE_TARGET    = 132  # Defined in RFC4684
BGP_SAFI_FLOWSPEC        = 133  # Defined in RFC8955
BGP_SAFI_L3VPN_FLOWSPEC  = 134  # Defined in RFC8955
BGP_SAFI_L3VPN_AUTO      = 140  # Proposed in draft_ietf_l3vpn_bgpvpn_auto
BGP_SAFI_RESERVED_255    = 255  # Defined in RFC4760

################
# BGP MESSAGES #
################

# BGP messages
BGP_BGP4MP_RESERVED      = 0  # Defined in RFC4271
BGP_BGP4MP_OPEN          = 1  # Defined in RFC4271
BGP_BGP4MP_UPDATE        = 2  # Defined in RFC4271
BGP_BGP4MP_NOTIFICATION  = 3  # Defined in RFC4271
BGP_BGP4MP_KEEPALIVE     = 4  # Defined in RFC4271
BGP_BGP4MP_ROUTE_REFRESH = 5  # Defined in RFC2918

##############
# BGP STATES #
##############

# BGP states
BGP_STATE_IDLE         = 1  # Defined in RFC6396
BGP_STATE_CONNECT      = 2  # Defined in RFC6396
BGP_STATE_ACTIVE       = 3  # Defined in RFC6396
BGP_STATE_OPEN_SENT    = 4  # Defined in RFC6396
BGP_STATE_OPEN_CONFIRM = 5  # Defined in RFC6396
BGP_STATE_ESTABLISHED  = 6  # Defined in RFC6396
BGP_STATE_CLEARING     = 7  # Quagga/FRR
BGP_STATE_DELETED      = 8  # Quagga/FRR

##############
# BGP PARAMS #
##############

# BGP parameters
BGP_PARAMS_RESERVED        = 0    # Defined in RFC5492
BGP_PARAMS_AUTHENTICATION  = 1    # Deprecated in RFC5492
BGP_PARAMS_CAPABILITIES    = 2    # Defined in RFC5492
BGP_PARAMS_EXTENDED_LENGTH = 255  # Defined in RFC9072

####################
# BGP CAPABILITIES #
####################

# BGP capabilities
BGP_CAPABILITY_RESERVED_0                 = 0    # Defined in RFC5492
BGP_CAPABILITY_BGP4MP                     = 1    # Defined in RFC2858
BGP_CAPABILITY_ROUTE_REFRESH              = 2    # Defined in RFC2918
BGP_CAPABILITY_OUTBOUND_FILTER            = 3    # Defined in RFC5291
BGP_CAPABILITY_MULTIPLE_ROUTES            = 4    # Deprecated in RFC8277
BGP_CAPABILITY_EXTENDED_NEXT_HOP          = 5    # Defined in RFC8950
BGP_CAPABILITY_BGP4MP_ET                  = 6    # Defined in RFC8654
BGP_CAPABILITY_BGPSEC                     = 7    # Defined in RFC8205
BGP_CAPABILITY_MULTIPLE_LABELS            = 8    # Defined in RFC8277
BGP_CAPABILITY_BGP_ROLE                   = 9    # Defined in RFC9234
BGP_CAPABILITY_GRACEFUL_RESTART           = 64   # Defined in RFC4724
BGP_CAPABILITY_AS4                        = 65   # Defined in RFC6793
BGP_CAPABILITY_UNKNOWN_66                 = 66   # Deprecated on 2003-03-06
BGP_CAPABILITY_DYNAMIC                    = 67   # Proposed in draft-ietf-idr-dynamic-cap
BGP_CAPABILITY_MULTISESSION               = 68   # Proposed in draft-ietf-idr-bgp-multisession
BGP_CAPABILITY_ADDPATH                    = 69   # Defined in RFC7911
BGP_CAPABILITY_ENHANCED_ROUTE_REFRESH     = 70   # Defined in RFC7313
BGP_CAPABILITY_LLGR                       = 71   # Proposed in draft-ietf-idr-long-lived-gr
BGP_CAPABILITY_POLICY_DISTRIBUTION        = 72   # Proposed in draft-ietf-idr-rpd-04
BGP_CAPABILITY_FQDN                       = 73   # Proposed in draft-walton-bgp-hostname-capability
BGP_CAPABILITY_BFD                        = 74   # Proposed in draft-ietf-idr-bgp-bfd-strict-mode
BGP_CAPABILITY_SOFTWARE_VERSION           = 75   # Proposed in draft-abraitis-bgp-version-capability
BGP_CAPABILITY_PRESTD_ROUTE_REFRESH       = 128  # Deprecated in RFC8810
BGP_CAPABILITY_PRESTD_POLICY_DISTRIBUTION = 129  # Deprecated in RFC8810
BGP_CAPABILITY_PRESTD_OUTBOUND_FILTER     = 130  # Deprecated in RFC8810
BGP_CAPABILITY_PRESTD_MULTISESSION        = 131  # Deprecated in RFC8810
BGP_CAPABILITY_PRESTD_FQDN                = 184  # Deprecated in RFC8810
BGP_CAPABILITY_PRESTD_OPERATIONAL         = 185  # Deprecated in RFC8810
BGP_CAPABILITY_RESERVED_255               = 255  # Defined in RFC8810

##############
# BGP ERRORS #
##############

# BGP error codes
BGP_ERROR_RESERVED           = 0  # Defined in RFC4271
BGP_ERROR_MESSAGE_HEADER     = 1  # Defined in RFC4271
BGP_ERROR_OPEN_MESSAGE       = 2  # Defined in RFC4271
BGP_ERROR_UPDATE_MESSAGE     = 3  # Defined in RFC4271
BGP_ERROR_HOLD_TIMER_EXPIRED = 4  # Defined in RFC4271
BGP_ERROR_FSM                = 5  # Defined in RFC4271
BGP_ERROR_CEASE              = 6  # Defined in RFC4271
BGP_ERROR_ROUTE_REFRESH      = 7  # Defined in RFC7313

# BGP message header error subcodes
BGP_ERROR_MESSAGE_HEADER_UNSPECIFIC                  = 0  # Defined in RFC4493
BGP_ERROR_MESSAGE_HEADER_CONNECTION_NOT_SYNCHRONIZED = 1  # Defined in RFC4271
BGP_ERROR_MESSAGE_HEADER_BAD_MESSAGE_LENGTH          = 2  # Defined in RFC4271
BGP_ERROR_MESSAGE_HEADER_BAD_MESSAGE_TYPE            = 3  # Defined in RFC4271

# BGP OPEN message error subcodes
BGP_ERROR_OPEN_MESSAGE_UNSPECIFIC                     = 0   # Defined in RFC4493
BGP_ERROR_OPEN_MESSAGE_UNSUPPORTED_VERSION_NUMBER     = 1   # Defined in RFC4271
BGP_ERROR_OPEN_MESSAGE_BAD_PEER_AS                    = 2   # Defined in RFC4271
BGP_ERROR_OPEN_MESSAGE_BAD_BGP_IDENTIFIER             = 3   # Defined in RFC4271
BGP_ERROR_OPEN_MESSAGE_UNSUPPORTED_OPTIONAL_PARAMETER = 4   # Defined in RFC4271
BGP_ERROR_OPEN_MESSAGE_AUTHENTICATION_FAILURE         = 5   # Deprecated in RFC4271
BGP_ERROR_OPEN_MESSAGE_UNACCEPTABLE_HOLD_TIME         = 6   # Defined in RFC4271
BGP_ERROR_OPEN_MESSAGE_UNSUPPORTED_CAPABILITY         = 7   # Defined in RFC5492
BGP_ERROR_OPEN_MESSAGE_DEPRECATED_8                   = 8   # Deprecated in RFC9234
BGP_ERROR_OPEN_MESSAGE_DEPRECATED_9                   = 9   # Deprecated in RFC9234
BGP_ERROR_OPEN_MESSAGE_DEPRECATED_10                  = 10  # Deprecated in RFC9234
BGP_ERROR_OPEN_MESSAGE_ROLE_MISMATCH                  = 11  # Defined in RFC9234

# BGP UPDATE message error subcodes
BGP_ERROR_UPDATE_MESSAGE_UNSPECIFIC                       = 0   # Defined in RFC4493
BGP_ERROR_UPDATE_MESSAGE_MALFORMED_ATTRIBUTE_LIST         = 1   # Defined in RFC4271
BGP_ERROR_UPDATE_MESSAGE_UNRECOGNIZED_WELLKNOWN_ATTRIBUTE = 2   # Defined in RFC4271
BGP_ERROR_UPDATE_MESSAGE_MISSING_WELLKNOWN_ATTRIBUTE      = 3   # Defined in RFC4271
BGP_ERROR_UPDATE_MESSAGE_INVALID_ATTRIBUTE_FLAGS          = 4   # Defined in RFC4271
BGP_ERROR_UPDATE_MESSAGE_INVALID_ATTRIBUTE_LENGTH         = 5   # Defined in RFC4271
BGP_ERROR_UPDATE_MESSAGE_INVALID_ORIGIN_ATTRIBUTE         = 6   # Defined in RFC4271
BGP_ERROR_UPDATE_MESSAGE_AS_ROUTING_LOOP                  = 7   # Deprecated in RFC4271
BGP_ERROR_UPDATE_MESSAGE_INVALID_NEXTHOP_ATTRIBUTE        = 8   # Defined in RFC4271
BGP_ERROR_UPDATE_MESSAGE_INVALID_OPTIONAL_ATTRIBUTE       = 9   # Defined in RFC4271
BGP_ERROR_UPDATE_MESSAGE_INVALID_NETWORK_FIELD            = 10  # Defined in RFC4271
BGP_ERROR_UPDATE_MESSAGE_MALFORMED_AS_PATH                = 11  # Defined in RFC4271

# BGP finite state machine error subcodes
BGP_ERROR_FSM_UNSPECIFIED                              = 0  # Defined in RFC6608
BGP_ERROR_FSM_UNEXPECTED_MESSAGE_IN_OPEN_SENT_STATE    = 1  # Defined in RFC6608
BGP_ERROR_FSM_UNEXPECTED_MESSAGE_IN_OPEN_CONFIRM_STATE = 2  # Defined in RFC6608
BGP_ERROR_FSM_UNEXPECTED_MESSAGE_IN_ESTABLISHED_STATE  = 3  # Defined in RFC6608

# BGP cease error subcodes
BGP_ERROR_CEASE_RESERVED                        = 0   # Defined in RFC4486
BGP_ERROR_CEASE_MAX_NUMBER_PREFIXES_REACHED     = 1   # Defined in RFC4486
BGP_ERROR_CEASE_ADMINISTRATIVE_SHUTDOWN         = 2   # Defined in RFC9003
BGP_ERROR_CEASE_PEER_DECONFIGURED               = 3   # Defined in RFC4486
BGP_ERROR_CEASE_ADMINISTRATIVE_RESET            = 4   # Defined in RFC9003
BGP_ERROR_CEASE_CONNECTION_REJECTED             = 5   # Defined in RFC4486
BGP_ERROR_CEASE_OTHER_CONFIGURATION_CHANGE      = 6   # Defined in RFC4486
BGP_ERROR_CEASE_CONNECTION_COLLISION_RESOLUTION = 7   # Defined in RFC4486
BGP_ERROR_CEASE_OUT_OF_RESOURCES                = 8   # Defined in RFC4486
BGP_ERROR_CEASE_HARD_RESET                      = 9   # Defined in RFC8538
BGP_ERROR_CEASE_BFD_DOWN                        = 10  # Defined in RFC9384

# BGP route refresh error subcodes
BGP_ERROR_ROUTE_REFRESH_RESERVED               = 0  # Defined in RFC7313
BGP_ERROR_ROUTE_REFRESH_INVALID_MESSAGE_LENGTH = 1  # Defined in RFC7313

#######################
# BGP PATH ATTRIBUTES #
#######################

# BGP path attribute types
BGP_PATH_ATTR_RESERVED                  = 0    # Defined in RFC4271
BGP_PATH_ATTR_ORIGIN                    = 1    # Defined in RFC4271
BGP_PATH_ATTR_AS_PATH                   = 2    # Defined in RFC4271
BGP_PATH_ATTR_NEXT_HOP                  = 3    # Defined in RFC4271
BGP_PATH_ATTR_MULTI_EXIT_DISC           = 4    # Defined in RFC4271
BGP_PATH_ATTR_LOCAL_PREF                = 5    # Defined in RFC4271
BGP_PATH_ATTR_ATOMIC_AGGREGATE          = 6    # Defined in RFC4271
BGP_PATH_ATTR_AGGREGATOR                = 7    # Defined in RFC4271
BGP_PATH_ATTR_COMMUNITIES               = 8    # Defined in RFC1997
BGP_PATH_ATTR_ORIGINATOR_ID             = 9    # Defined in RFC4456
BGP_PATH_ATTR_CLUSTER_LIST              = 10   # Defined in RFC4456
BGP_PATH_ATTR_DPA                       = 11   # Deprecated in RFC6938
BGP_PATH_ATTR_ADVERTISER                = 12   # Deprecated in RFC6938
BGP_PATH_ATTR_RCID_CLUSTER_ID           = 13   # Deprecated in RFC6938
BGP_PATH_ATTR_MP_REACH_NLRI             = 14   # Defined in RFC4760
BGP_PATH_ATTR_MP_UNREACH_NLRI           = 15   # Defined in RFC4760
BGP_PATH_ATTR_EXTENDED_COMMUNITIES      = 16   # Defined in RFC4360
BGP_PATH_ATTR_AS4_PATH                  = 17   # Defined in RFC6793
BGP_PATH_ATTR_AS4_AGGREGATOR            = 18   # Defined in RFC6793
BGP_PATH_ATTR_SAFI_SPECIFIC             = 19   # Proposed in draft-kapoor-nalawade-idr-bgp-ssa
BGP_PATH_ATTR_CONNECTOR                 = 20   # Deprecated in RFC6037
BGP_PATH_ATTR_AS_PATHLIMIT              = 21   # Proposed in draft-ietf-idr-as-pathlimit
BGP_PATH_ATTR_PMSI_TUNNEL               = 22   # Defined in RFC6514
BGP_PATH_ATTR_TUNNEL_ENCAPSULATION      = 23   # Defined in RFC5512
BGP_PATH_ATTR_TRAFFIC_ENGINEERING       = 24   # Defined in RFC5543
BGP_PATH_ATTR_IPV6_EXTENDED_COMMUNITIES = 25   # Defined in RFC5701
BGP_PATH_ATTR_AIGP                      = 26   # Defined in RFC7311
BGP_PATH_ATTR_PE_DISTINGUISHER_LABELS   = 27   # Defined in RFC6514
BGP_PATH_ATTR_ENTROPY_LABEL_CAPABILITY  = 28   # Deprecated in RFC7447
BGP_PATH_ATTR_LS                        = 29   # Defined in RFC7752
BGP_PATH_ATTR_VENDOR_30                 = 30   # Deprecated in RFC8093
BGP_PATH_ATTR_VENDOR_31                 = 31   # Deprecated in RFC8093
BGP_PATH_ATTR_LARGE_COMMUNITIES         = 32   # Defined in RFC8092
BGP_PATH_ATTR_BGPSEC_PATH               = 33   # Defined in RFC8205
BGP_PATH_ATTR_COMMUNITY_CONTAINER       = 34   # Proposed in draft-ietf-idr-wide-bgp-communities
BGP_PATH_ATTR_ONLY_TO_CUSTOMER          = 35   # Proposed in draft-ietf-idr-bgp-open-policy
BGP_PATH_ATTR_DOMAIN_PATH               = 36   # Proposed in draft-ietf-bess-evpn-ipvpn-interworking
BGP_PATH_ATTR_SFP                       = 37   # Defined in RFC9015
BGP_PATH_ATTR_BFD_DISCRIMINATOR         = 38   # Defined in RFC9026
BGP_PATH_ATTR_NHC_TMP                   = 39   # Proposed in draft-ietf-idr-entropy-label
BGP_PATH_ATTR_PREFIX_SID                = 40   # Defined in RFC8669
BGP_PATH_ATTR_ATTR_SET                  = 128  # Defined in RFC6368
BGP_PATH_ATTR_VENDOR_129                = 129  # Deprecated in RFC8093
BGP_PATH_ATTR_VENDOR_241                = 241  # Deprecated in RFC8093
BGP_PATH_ATTR_VENDOR_242                = 242  # Deprecated in RFC8093
BGP_PATH_ATTR_VENDOR_243                = 243  # Deprecated in RFC8093
BGP_PATH_ATTR_RESERVED_FOR_DEV          = 255  # Defined in RFC2042

# AS path segment types
BGP_PATH_ATTR_AS_PATH_SEGMENT_SET             = 1  # Defined in RFC4271
BGP_PATH_ATTR_AS_PATH_SEGMENT_SEQUENCE        = 2  # Defined in RFC4271
BGP_PATH_ATTR_AS_PATH_SEGMENT_CONFED_SEQUENCE = 3  # Defined in RFC5065
BGP_PATH_ATTR_AS_PATH_SEGMENT_CONFED_SET      = 4  # Defined in RFC5065

# AS4 path transition
BGP_PATH_ATTR_AS4_PATH_AS_TRANS = 23456  # Defined in RFC6793

# Origin types
BGP_PATH_ATTR_ORIGIN_IGP        = 0  # Defined in RFC4271
BGP_PATH_ATTR_ORIGIN_EGP        = 1  # Defined in RFC4271
BGP_PATH_ATTR_ORIGIN_INCOMPLETE = 2  # Defined in RFC4271
